"""Get details about a security group."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Get details about a security group."""

    mgr = SoftLayer.NetworkManager(env.client)

    secgroup = mgr.get_securitygroup(identifier)

    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'

    table.add_row(['id', secgroup['id']])
    table.add_row(['name', secgroup.get('name') or formatting.blank()])
    table.add_row(['description',
                   secgroup.get('description') or formatting.blank()])

    rule_table = formatting.Table(['id', 'remoteIp', 'remoteGroupId',
                                   'direction', 'ethertype', 'portRangeMin',
                                   'portRangeMax', 'protocol'])
    for rule in secgroup.get('rules', []):
        rg_id = rule.get('remoteGroup', {}).get('id') or formatting.blank()
        port_min = rule.get('portRangeMin')
        port_max = rule.get('portRangeMax')
        if port_min is None:
            port_min = formatting.blank()
        if port_max is None:
            port_max = formatting.blank()
        rule_table.add_row([rule['id'],
                            rule.get('remoteIp') or formatting.blank(),
                            rule.get('remoteGroupId', rg_id),
                            rule['direction'],
                            rule.get('ethertype') or formatting.blank(),
                            port_min,
                            port_max,
                            rule.get('protocol') or formatting.blank()])

    table.add_row(['rules', rule_table])

    vsi_table = formatting.Table(['id', 'hostname', 'interface', 'ipAddress'])

    for binding in secgroup.get('networkComponentBindings', []):
        try:
            vsi = binding['networkComponent']['guest']
            vsi_id = vsi['id']
            hostname = vsi['hostname']
            interface = ('PRIVATE' if binding['networkComponent']['port'] == 0
                         else 'PUBLIC')
            ip_address = (vsi['primaryBackendIpAddress']
                          if binding['networkComponent']['port'] == 0
                          else vsi['primaryIpAddress'])
        except KeyError:
            vsi_id = "N/A"
            hostname = "Not enough permission to view"
            interface = "N/A"
            ip_address = "N/A"
        vsi_table.add_row([vsi_id, hostname, interface, ip_address])

    table.add_row(['servers', vsi_table])

    env.fout(table)



"""Cancel an existing file storage volume."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('volume-id')
@click.option('--reason', help="An optional reason for cancellation")
@click.option('--immediate', is_flag=True,
              help="Cancels the file storage volume immediately instead of on the billing anniversary")
@click.option('--force', default=False, is_flag=True, help="Force modify")
@environment.pass_env
def cli(env, volume_id, reason, immediate, force):
    """Cancel an existing file storage volume.

    EXAMPLE::
        slcli file volume-cancel 12345678 --immediate -f
        This command cancels volume with ID 12345678 immediately and without asking for confirmation.
    """

    file_storage_manager = SoftLayer.FileStorageManager(env.client)

    if not force:
        if not (env.skip_confirmations or formatting.no_going_back(volume_id)):
            raise exceptions.CLIAbort('Aborted.')

    cancelled = file_storage_manager.cancel_file_volume(volume_id, reason, immediate)

    if cancelled:
        if immediate:
            click.echo(f'File volume with id {volume_id} has been marked for immediate cancellation')
        else:
            click.echo(f'File volume with id {volume_id} has been marked for cancellation')
    else:
        click.echo(f'Unable to cancle file volume {volume_id}')



"""View a quote"""
# :license: MIT, see LICENSE for more details.
import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers import ordering
from SoftLayer.utils import lookup


@click.command(cls=SLCommand)
@click.argument('quote')
@environment.pass_env
def cli(env, quote):
    """View a quote"""

    manager = ordering.OrderingManager(env.client)
    result = manager.get_quote_details(quote)

    package = result['order']['items'][0]['package']
    title = f"{result.get('name')} - Package: {package['keyName']}, Id {package['id']}"
    table = formatting.Table([
        'Category', 'Description', 'Quantity', 'Recurring', 'One Time'
    ], title=title)
    table.align['Category'] = 'l'
    table.align['Description'] = 'l'

    items = lookup(result, 'order', 'items')
    for item in items:
        table.add_row([
            item.get('categoryCode'),
            item.get('description'),
            item.get('quantity'),
            item.get('recurringFee'),
            item.get('oneTimeFee')
        ])

    env.fout(table)



STAAS_TEST_VOLUME = {
    'accountId': 1234,
    'activeTransactions': None,
    'activeTransactionCount': 0,
    'billingItem': {
        'activeChildren': [{
            'categoryCode': 'storage_snapshot_space',
            'id': 125,
            'cancellationDate': '',
        }],
        'cancellationDate': '',
        'categoryCode': 'storage_as_a_service',
        'hourlyFlag': None,
        'id': 454,
        'location': {'id': 449500}
    },
    'capacityGb': 500,
    'hasEncryptionAtRest': 1,
    'id': 102,
    'iops': 1000,
    'lunId': 2,
    'osType': {'keyName': 'LINUX'},
    'originalVolumeSize': '500',
    'parentVolume': {'snapshotSizeBytes': 1024},
    'provisionedIops': '1000',
    'replicationPartnerCount': 0,
    'schedules': [{
        'id': 978,
        'type': {'keyname': 'SNAPSHOT_WEEKLY'},
    }],
    'serviceResource': {'datacenter': {'id': 449500, 'name': 'dal05'}},
    'serviceResourceBackendIpAddress': '10.1.2.3',
    'snapshotCapacityGb': '10',
    'staasVersion': '2',
    'storageTierLevel': 'READHEAVY_TIER',
    'storageType': {'keyName': 'ENDURANCE_BLOCK_STORAGE'},
    'username': 'duplicatable_volume_username'
}

getObject = {
    'accountId': 1234,
    'activeTransactionCount': 1,
    'activeTransactions': [{
        'transactionStatus': {'friendlyName': 'This is a buffer time in which the customer may cancel the server'}
    }],
    'allowedHardware': [{
        'allowedHost': {
            'credential': {'username': 'joe', 'password': '12345'},
            'name': 'test-server',
        },
        'domain': 'example.com',
        'hostname': 'test-server',
        'id': 1234,
        'primaryBackendIpAddress': '10.0.0.2',
    }],
    'allowedIpAddresses': [{
        'allowedHost': {
            'credential': {'username': 'joe', 'password': '12345'},
            'name': 'test-server',
        },
        'id': 1234,
        'ipAddress': '10.0.0.1',
        'note': 'backend ip',
    }],
    'allowedSubnets': [{
        'allowedHost': {
            'credential': {'username': 'joe', 'password': '12345'},
            'name': 'test-server',
        },
        'cidr': '24',
        'id': 1234,
        'networkIdentifier': '10.0.0.1',
        'note': 'backend subnet',
    }],
    'allowedVirtualGuests': [{
        'allowedHost': {
            'credential': {'username': 'joe', 'password': '12345'},
            'name': 'test-server',
        },
        'domain': 'example.com',
        'hostname': 'test-server',
        'id': 1234,
        'primaryBackendIpAddress': '10.0.0.1',
    }],
    'billingItem': {
        'activeChildren': [{
            'cancellationDate': '',
            'categoryCode': 'storage_snapshot_space',
            'id': 123,
        }],
        'cancellationDate': '',
        'categoryCode': 'storage_service_enterprise',
        'id': 449,
        'location': {'id': 449500}
    },
    'bytesUsed': 0,
    'capacityGb': 20,
    'createDate': '2015:50:15-04:00',
    'fileNetworkMountAddress': '127.0.0.1:/TEST',
    'guestId': '',
    'hardwareId': '',
    'hasEncryptionAtRest': 0,
    'hostId': '',
    'id': 100,
    'iops': 1000,
    'lunId': 2,
    'nasType': 'ISCSI',
    'notes': """{'status': 'available'}""",
    'originalSnapshotName': 'test-original-snapshot-name',
    'originalVolumeName': 'test-original-volume-name',
    'originalVolumeSize': '20',
    'osType': {'keyName': 'LINUX'},
    'parentVolume': {'snapshotSizeBytes': 1024},
    'password': '',
    'provisionedIops': '1000',
    'replicationPartnerCount': 1,
    'replicationPartners': [{
        'createDate': '2017:50:15-04:00',
        'id': 1784,
        'nasType': 'ISCSI_REPLICANT',
        'replicationSchedule': {'type': {'keyname': 'REPLICATION_HOURLY'}},
        'serviceResource': {'datacenter': {'name': 'wdc01'}},
        'serviceResourceBackendIpAddress': '10.3.174.79',
        'username': 'TEST_REP_1',
    }, {
        'createDate': '2017:50:15-04:00',
        'id': 1785,
        'nasType': 'ISCSI_REPLICANT',
        'replicationSchedule': {'type': {'keyname': 'REPLICATION_DAILY'}},
        'serviceResource': {'datacenter': {'name': 'dal01'}},
        'serviceResourceBackendIpAddress': '10.3.177.84',
        'username': 'TEST_REP_2',
    }],
    'replicationStatus': 'Replicant Volume Provisioning has completed.',
    'schedules': [
        {
            'id': 978,
            'type': {'keyname': 'SNAPSHOT_WEEKLY'},
            'properties': [
                {'type': {'keyname': 'MINUTE'}, 'value': '30'},
            ]
        },
        {
            'id': 988,
            'type': {'keyname': 'REPLICATION_INTERVAL'},
            'properties': [
                {'type': {'keyname': 'MINUTE'}, 'value': '-1'},
            ]
        }
    ],
    'serviceProviderId': 1,
    'serviceResource': {'datacenter': {'id': 449500, 'name': 'dal05'},
                        'name': 'Cleversafe - US Region',
                        'type': {
                            'type': 'CLEVERSAFE_SVC_API'
                        }},
    'serviceResourceBackendIpAddress': '10.1.2.3',
    'serviceResourceName': 'Storage Type 01 Aggregate staaspar0101_pc01',
    'snapshotCapacityGb': '10',
    'staasVersion': '1',
    'storageTierLevel': 'READHEAVY_TIER',
    'storageType': {'keyName': 'ENDURANCE_STORAGE'},
    'username': 'username',
    'dependentDuplicate': 1,
}

getSnapshots = [{
    'id': 470,
    'notes': 'unit_testing_note',
    'snapshotCreationTimestamp': '2016-07-06T07:41:19-05:00',
    'snapshotSizeBytes': '42',
}]

getReplicationPartners = [{
    'id': 1784,
    'accountId': 3000,
    'capacityGb': 20,
    'username': 'TEST_REP_1',
    'serviceResourceBackendIpAddress': '10.3.174.79',
    'nasType': 'ISCSI_REPLICANT',
    'hostId': None,
    'guestId': None,
    'hardwareId': None,
    'createDate': '2017:50:15-04:00',
    'serviceResource': {'datacenter': {'name': 'wdc01'}},
    'replicationSchedule': {'type': {'keyname': 'REPLICATION_HOURLY'}},
}, {
    'id': 1785,
    'accountId': 3001,
    'capacityGb': 20,
    'username': 'TEST_REP_2',
    'serviceResourceBackendIpAddress': '10.3.177.84',
    'nasType': 'ISCSI_REPLICANT',
    'hostId': None,
    'guestId': None,
    'hardwareId': None,
    'createDate': '2017:50:15-04:00',
    'serviceResource': {'datacenter': {'name': 'dal01'}},
    'replicationSchedule': {'type': {'keyname': 'REPLICATION_DAILY'}},
}]

getValidReplicationTargetDatacenterLocations = [{
    'id': 12345,
    'longName': 'Dallas 05',
    'name': 'dal05'
}]

listVolumeSchedules = [
    {
        'id': 978,
        'type': {'keyname': 'SNAPSHOT_WEEKLY'},
        'properties': [{'type': {'keyname': 'MINUTE'}, 'value': '30'}]
    },
    {
        'id': 988,
        'type': {'keyname': 'REPLICATION_INTERVAL'},
        'properties': [{'type': {'keyname': 'MINUTE'}, 'value': '-1'}]
    }
]

deleteObject = True
editObject = True
allowAccessFromHostList = True
removeAccessFromHostList = True
failoverToReplicant = True
failbackFromReplicant = True
restoreFromSnapshot = True
disasterRecoveryFailoverToReplicant = True

createSnapshot = {
    'id': 449
}

enableSnapshots = True
disableSnapshots = True

getVolumeCountLimits = [{
    "datacenterName": "global",
    "maximumAvailableCount": 700,
    "provisionedCount": 2632
},
    {
        "maximumAvailableCount": 50,
        "provisionedCount": 2632
    },
    {
        "datacenterName": "dal13",
        "maximumAvailableCount": 52,
        "provisionedCount": 30
    },
]

refreshDuplicate = {
    'dependentDuplicate': 1
}

convertCloneDependentToIndependent = {
    'dependentDuplicate': 1
}

getDuplicateConversionStatus = {
    'activeConversionStartTime': '2023-01-01',
    'deDuplicateConversionPercentage': 50,
    'volumeUsername': 'testUser'
}

BLOCK_LIST_ISSUES_1732 = {
    'capacityGb': 16000,
    'id': 167214314,
    'notes': 'test',
    'username': 'SL02SEL307608-60',
    'activeTransactionCount': 0,
    'replicationPartnerCount': 1,
    'lunId': '1',
    'parentVolume': {
        'accountId': 307608,
        'capacityGb': 16000,
        'createDate': '2020-09-02T14:55:32-06:00',
        'guestId': None,
        'hardwareId': None,
        'hostId': None,
        'id': 167214302,
        'nasType': 'NAS_CONTAINER',
        'serviceProviderId': 1,
        'storageTypeId': '3',
        'upgradableFlag': True,
        'username': 'SL02SEVC307608_60',
        'serviceResourceBackendIpAddress': 'fsf-mex0101b-fz.service.softlayer.com',
        'serviceResourceName': 'Storage Type 02 Aggregate stxf-mex0101b',
        'snapshotSizeBytes': '831488'
    },
    'provisionedIops': '4096',
    'replicationPartners': [{
        'id': 167236648,
        'username': 'SL02SEL307608_60_REP_1',
        'replicationSchedule': {
            'active': 1,
            'createDate': '2020-09-02T17:10:04-06:00',
            'id': 542062,
            'modifyDate': None,
            'name': 'SL02SEVC307608_60_WEEKLY',
            'partnershipId': None,
            'typeId': 34,
            'volumeId': 167214302,
            'type': {
                'keyname': 'REPLICATION_WEEKLY'
            }
        },
        'serviceResource': {
            'backendIpAddress': 'fsf-sao0102c-fz.service.softlayer.com',
            'id': 30650,
            'name': 'Storage Type 02 Aggregate stxf-sao0102c',
            'datacenter': {
                'name': 'sao01'
            },
            'type': {
                'type': 'NETAPP_STOR_AGGR'
            }
        },
        'serviceResourceBackendIpAddress': '10.200.14.72'
    }],
    'replicationStatus': 'FAILBACK_COMPLETED',
    'serviceResource': {
        'backendIpAddress': 'fsf-mex0101b-fz.service.softlayer.com',
        'id': 9291,
        'name': 'Storage Type 02 Aggregate stxf-mex0101b',
        'datacenter': {
            'name': 'mex01'
        }
    },
    'serviceResourceBackendIpAddress': '10.2.190.99',
    'snapshotCapacityGb': '1000',
    'storageTierLevel': 'LOW_INTENSITY_TIER',
    'storageType': {
        'keyName': 'ENDURANCE_BLOCK_STORAGE'
    }
}


FILE_DETAIL_ISSUE2154 = {
    "capacityGb": 150,
    "id": 609491933,
    "username": "SL02SV1414935_187",
    "activeTransactionCount": 0,
    "replicationPartnerCount": 1,
    "fileNetworkMountAddress": "fsf-natestdal0505-fcb-fz.service.softlayer.com:/SL02SV1414935_187/data01",
    "originalVolumeSize": "20",
    "provisionedIops": "2000",
    "replicationStatus": "FAILOVER_COMPLETED",
    "serviceResourceBackendIpAddress": "fsf-natestdal0505-fcb-fz.service.softlayer.com",
    "snapshotCapacityGb": "5",
    "activeTransactions": [
        {
            "createDate": "",
            "elapsedSeconds": 111763,
            "guestId": "",
            "hardwareId": "",
            "id": "",
            "modifyDate": "",
            "statusChangeDate": "",
            "transactionGroup": {
                "name": "Volume Modification"
            },
            "transactionStatus": {
                "friendlyName": "In Progress"
            }
        }
    ],
    "parentVolume": {
        "accountId": 1414935,
        "capacityGb": 120,
        "createDate": "2024-05-16T02:28:02-05:00",
        "guestId": "",
        "hardwareId": "",
        "hostId": "",
        "id": 609491967,
        "nasType": "SNAPSHOT",
        "notes": "vol_duplicate_snapshot_2024-05-16_0228",
        "serviceProviderId": 1,
        "storageTypeId": "16",
        "upgradableFlag": True,
        "username": "SL02SV1414935_187",
        "serviceResourceBackendIpAddress": "fsf-natestdal0505-fcb-fz.service.softlayer.com",
        "serviceResourceName": "Storage Type 02 Aggregate natestdal0505-fc-d",
        "snapshotSizeBytes": "0"
    },
    "replicationPartners": [
        {
            "id": 609491945,
            "username": "SL02SV1414935_187_REP_1",
            "serviceResourceBackendIpAddress": "fsf-natestdal0505-ffb-fz.service.softlayer.com",
            "replicationSchedule": {
                "active": 1,
                "createDate": "2024-05-16T01:20:19-05:00",
                "id": 666339,
                "modifyDate": "",
                "name": "SL02SV1414935_187_HOURLY_REP",
                "partnershipId": "",
                "typeId": 32,
                "volumeId": 609491933,
                "type": {
                    "keyname": "REPLICATION_HOURLY"
                }
            },
            "serviceResource": {
                "backendIpAddress": "fsf-natestdal0505-ffb-fz.service.softlayer.com",
                "id": 57365,
                "name": "Storage Type 02 Aggregate natestdal0505-ff-d",
                "datacenter": {
                    "name": "dal10"
                },
                "type": {
                    "type": "NETAPP_STOR_AGGR"
                }
            }
        }
    ],
    "serviceResource": {
        "backendIpAddress": "fsf-natestdal0505-fcb-fz.service.softlayer.com",
        "id": 52292,
        "name": "Storage Type 02 Aggregate natestdal0505-fc-d",
        "attributes": [
            {
                "value": "2",
                "attributeType": {
                    "keyname": "STAAS_VERSION"
                }
            }
        ],
        "datacenter": {
            "name": "lon02"
        },
        "type": {
            "type": "NETAPP_STOR_AGGR"
        }
    },
    "storageType": {
        "keyName": "PERFORMANCE_FILE_STORAGE"
    }
}



"""Export an image."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.argument('uri')
@click.option('--ibm-api-key',
              default=None,
              help="The IBM Cloud API Key with access to IBM Cloud Object "
                   "Storage instance. For help creating this key see "
                   "https://cloud.ibm.com/docs/cloud-object-storage?"
                   "topic=cloud-object-storage-iam-overview#iam-overview"
                   "-service-id-api-key "
              )
@environment.pass_env
def cli(env, identifier, uri, ibm_api_key):
    """Export an image to object storage.

    The URI for an object storage object (.vhd/.iso file) of the format:
    swift://<objectStorageAccount>@<cluster>/<container>/<objectPath>
    or cos://<regionName>/<bucketName>/<objectPath> if using IBM Cloud
    Object Storage
    """

    image_mgr = SoftLayer.ImageManager(env.client)
    image_id = helpers.resolve_id(image_mgr.resolve_ids, identifier, 'image')
    result = image_mgr.export_image_to_uri(image_id, uri, ibm_api_key)

    if not result:
        raise exceptions.CLIAbort("Failed to export Image")



getAllObjects = [
    {'id': 100,
     'name': 'secgroup1',
     'description': 'Securitygroup1'},
    {'id': 104,
     'name': 'secgroup2'},
    {'id': 110}
]

getRules = [
    {'id': 100,
     'direction': 'egress',
     'ethertype': 'IPv4'}
]

guest_dict = {'id': 5000,
              'hostname': 'test',
              'primaryBackendIpAddress': '10.3.4.5',
              'primaryIpAddress': '169.23.123.43'}

getObject = {
    'id': 100,
    'name': 'secgroup1',
    'description': 'Securitygroup1',
    'networkComponentBindings': [{'networkComponentId': 1000,
                                  'networkComponent': {'id': 1000,
                                                       'port': 0,
                                                       'guest': guest_dict}},
                                 {'networkComponentId': 1001,
                                  'networkComponent': {'id': 1001,
                                                       'port': 1,
                                                       'guest': guest_dict}}],
    'rules': getRules
}

createObject = {'id': 100,
                'name': 'secgroup1',
                'description': 'Securitygroup1',
                'createDate': '2017-05-05T12:44:43-06:00'}
editObject = True
deleteObject = True
addRules = {"requestId": "addRules",
            "rules": "[{'direction': 'ingress', "
                     "'portRangeMax': '', "
                     "'portRangeMin': '', "
                     "'ethertype': 'IPv4', "
                     "'securityGroupId': 100, "
                     "'remoteGroupId': '', "
                     "'id': 100}]"}
editRules = {'requestId': 'editRules'}
removeRules = {'requestId': 'removeRules'}
attachNetworkComponents = {'requestId': 'interfaceAdd'}
detachNetworkComponents = {'requestId': 'interfaceRemove'}



"""Cancels a firewall."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--firewall-type', required=True, show_default=True, default='vlan',
              type=click.Choice(['vlan', 'server'], case_sensitive=False),
              help='Firewall type.')
@click.option('-f', '--force',  default=False, is_flag=True, help="Force cancel firewall of the server")
@environment.pass_env
def cli(env, identifier, firewall_type, force):
    """Cancels a firewall."""

    mgr = SoftLayer.FirewallManager(env.client)
    if not force:
        if not (env.skip_confirmations or
                formatting.confirm("This action will cancel a firewall from your account. Continue?")):
            raise exceptions.CLIAbort('Aborted.')

    if firewall_type == 'server':
        mgr.cancel_firewall(identifier, dedicated=False)
    else:
        mgr.cancel_firewall(identifier, dedicated=True)

    env.fout('Firewall with id %s is being cancelled!' % identifier)



"""Delete zone."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('zone')
@environment.pass_env
def cli(env, zone):
    """Delete zone.

    Example::

        slcli dns zone-delete ibm.com
        This command deletes a zone that is named ibm.com
"""

    manager = SoftLayer.DNSManager(env.client)
    zone_id = helpers.resolve_id(manager.resolve_ids, zone, name='zone')

    if not (env.skip_confirmations or formatting.no_going_back(zone)):
        raise exceptions.CLIAbort("Aborted.")

    manager.delete_zone(zone_id)



"""Cancel a snapshot space subscription."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('volume-id')
@click.option('--reason', help="An optional reason for cancellation.")
@click.option('--immediate',
              is_flag=True,
              help="Cancels the snapshot space immediately instead "
                   "of on the billing anniversary.")
@click.option('--force',  default=False, is_flag=True, help="Force modify")
@environment.pass_env
def cli(env, volume_id, reason, immediate, force):
    """Cancel existing snapshot space for a given volume."""

    block_storage_manager = SoftLayer.BlockStorageManager(env.client)

    if not (env.skip_confirmations or formatting.no_going_back(volume_id)):
        raise exceptions.CLIAbort('Aborted')

    if not force:
        if not (env.skip_confirmations or
                formatting.confirm("This action will incur charges on your account. Continue?")):
            raise exceptions.CLIAbort('Aborted')

    cancelled = block_storage_manager.cancel_snapshot_space(
        volume_id, reason, immediate)

    if cancelled:
        if immediate:
            click.echo('Block volume with id %s has been marked'
                       ' for immediate snapshot cancellation' % volume_id)
        else:
            click.echo('Block volume with id %s has been marked'
                       ' for snapshot cancellation' % volume_id)
    else:
        click.echo('Unable to cancel snapshot space for block volume %s'
                   % volume_id)



"""Cancel a dedicated host."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Cancel all virtual guests of the dedicated host immediately.

       Use the 'slcli vs cancel' command to cancel an specific guest
    """

    dh_mgr = SoftLayer.DedicatedHostManager(env.client)

    host_id = helpers.resolve_id(dh_mgr.resolve_ids, identifier, 'dedicated host')

    if not (env.skip_confirmations or formatting.no_going_back(host_id)):
        raise exceptions.CLIAbort('Aborted')

    table = formatting.Table(['id', 'server name', 'status'])

    result = dh_mgr.cancel_guests(host_id)

    if result:
        for status in result:
            table.add_row([
                status['id'],
                status['fqdn'],
                status['status']
            ])

        env.fout(table)
    else:
        click.secho('There is not any guest into the dedicated host %s' % host_id, fg='red')



"""Get details for a dedicated host."""
# :license: MIT, see LICENSE for more details.

import logging

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer import utils

LOGGER = logging.getLogger(__name__)


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--price', is_flag=True, help='Show associated prices')
@click.option('--guests', is_flag=True, help='Show guests on dedicated host')
@environment.pass_env
def cli(env, identifier, price=False, guests=False):
    """Get details for a dedicated host."""
    dhost = SoftLayer.DedicatedHostManager(env.client)

    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'

    result = dhost.get_host(identifier)
    result = utils.NestedDict(result)

    table.add_row(['id', result['id']])
    table.add_row(['name', result['name']])
    table.add_row(['cpu count', result['cpuCount']])
    table.add_row(['memory capacity', result['memoryCapacity']])
    table.add_row(['disk capacity', result['diskCapacity']])
    table.add_row(['create date', result['createDate']])
    table.add_row(['modify date', result['modifyDate']])
    table.add_row(['router id', result['backendRouter']['id']])
    table.add_row(['router hostname', result['backendRouter']['hostname']])
    table.add_row(['owner', formatting.FormattedItem(
        utils.lookup(result, 'billingItem', 'orderItem', 'order', 'userRecord', 'username') or formatting.blank(),)])

    if price:
        total_price = utils.lookup(result,
                                   'billingItem',
                                   'nextInvoiceTotalRecurringAmount') or 0
        total_price += sum(p['nextInvoiceTotalRecurringAmount']
                           for p
                           in utils.lookup(result,
                                           'billingItem',
                                           'children') or [])
        table.add_row(['price_rate', total_price])

    table.add_row(['guest count', result['guestCount']])
    if guests:
        guest_table = formatting.Table(['id', 'hostname', 'domain', 'uuid'])
        for guest in result['guests']:
            guest_table.add_row([
                guest['id'], guest['hostname'], guest['domain'], guest['uuid']])
        table.add_row(['guests', guest_table])

    table.add_row(['datacenter', result['datacenter']['name']])

    env.fout(table)



"""Delete a provisioning script"""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.managers.account import AccountManager as AccountManager


@click.command(cls=SLCommand)
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Delete a provisioning script"""

    manager = AccountManager(env.client)

    try:
        manager.delete_provisioning(identifier)
        click.secho("%s deleted successfully" % identifier, fg='green')
    except SoftLayer.SoftLayerAPIError as ex:
        click.secho("Failed to delete %s\n%s" % (identifier, ex), fg='red')



"""Cancel Network Vlan."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Cancel network VLAN."""

    mgr = SoftLayer.NetworkManager(env.client)

    if not (env.skip_confirmations or formatting.no_going_back(identifier)):
        raise exceptions.CLIAbort('Aborted')

    reasons = mgr.get_cancel_failure_reasons(identifier)
    if len(reasons) > 0:
        raise exceptions.CLIAbort(reasons)
    item = mgr.get_vlan(identifier).get('billingItem')
    if item:
        if mgr.cancel_item(item.get('id'),
                           True,
                           'Cancel by cli command',
                           'Cancel by cli command'):
            env.fout(f"VLAN {identifier} was cancelled.")
    else:
        raise exceptions.CLIAbort(
            "VLAN is an automatically assigned and free of charge VLAN,"
            " it will automatically be removed from your account when it is empty")



"""Remove the route of your Account Owned subnets."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Remove the route of your Account Owned subnets."""

    mgr = SoftLayer.NetworkManager(env.client)

    subnet = mgr.clear_route(identifier)

    if subnet:
        click.secho("The transaction to clear the route is created, routes will be updated in one or two minutes.")
    else:
        raise exceptions.CLIAbort('Failed to clear the route for the subnet.')



"""
    SoftLayer.formatting
    ~~~~~~~~~~~~~~~~~~~~
    Provider classes and helper functions to display output onto a command-line.

"""
# pylint: disable=E0202, consider-merging-isinstance, arguments-differ, keyword-arg-before-vararg
import collections
import csv
import io
import json
import os
import sys

import click
from rich import box
from rich.errors import NotRenderableError
from rich.table import Table as rTable

from SoftLayer.CLI import exceptions
from SoftLayer import utils

FALSE_VALUES = ['0', 'false', 'FALSE', 'no', 'False']


def format_output(data, fmt='table', theme=None):  # pylint: disable=R0911,R0912
    """Given some data, will format it for console output.

    :param data: One of: String, Table, FormattedItem, List, Tuple, SequentialOutput
    :param string fmt (optional): One of: table, raw, json, python
    """
    if fmt == 'json':
        return json.dumps(data, indent=4, cls=CLIJSONEncoder)
    elif fmt == 'jsonraw':
        return json.dumps(data, cls=CLIJSONEncoder)
    if fmt == 'csv':
        return csv_output_format(data)

    if isinstance(data, str) or isinstance(data, rTable):
        return data

    # responds to .prettytable()
    if hasattr(data, 'prettytable') and fmt in ('table', 'raw'):
        return format_prettytable(data, fmt, theme)

    # responds to .to_python()
    if hasattr(data, 'to_python'):
        if fmt == 'json':
            return json.dumps(format_output(data, fmt='python'), indent=4, cls=CLIJSONEncoder)
        elif fmt == 'jsonraw':
            return json.dumps(format_output(data, fmt='python'), cls=CLIJSONEncoder)
        elif fmt == 'python':
            return data.to_python()

    # responds to .formatted
    if hasattr(data, 'formatted'):
        if fmt == 'table':
            return data.formatted

    # responds to .separator
    if hasattr(data, 'separator'):
        output = [format_output(d, fmt=fmt) for d in data if d]
        return str(SequentialOutput(data.separator, output))

    # is iterable
    if isinstance(data, list) or isinstance(data, tuple):
        output = [format_output(d, fmt=fmt) for d in data]
        if fmt == 'python':
            return output
        return output

    # fallback, convert this odd object to a string
    return str(data)


def format_prettytable(table, fmt='table', theme=None):
    """Converts SoftLayer.CLI.formatting.Table instance to a prettytable."""
    for i, row in enumerate(table.rows):
        for j, item in enumerate(row):
            # Issue when adding items that evaulate to None (like empty lists) for Rich Tables
            # so we just cast those to a str
            if item:
                table.rows[i][j] = format_output(item)
            else:
                table.rows[i][j] = str(item)
    ptable = table.prettytable(fmt, theme)
    return ptable


def mb_to_gb(megabytes):
    """Converts number of megabytes to a FormattedItem in gigabytes.

    :param int megabytes: number of megabytes
    """
    return FormattedItem(megabytes, "%dG" % (float(megabytes) / 1024))


def b_to_gb(_bytes):
    """Converts number of bytes to a FormattedItem in gigabytes.

    :param int _bytes: number of bytes
    """
    return FormattedItem(_bytes, "%.2fG" % (float(_bytes) / 1024 / 1024 / 1024))


def gb(gigabytes):  # pylint: disable=C0103
    """Converts number of gigabytes to a FormattedItem in gigabytes.

    :param int gigabytes: number of gigabytes
    """
    return FormattedItem(int(float(gigabytes)) * 1024, "%dG" % int(float(gigabytes)))


def convert_sizes(value, unit='GB', round_result=False):
    """Converts a data storage value to an appropriate unit.

    :param str value: The value to convert.
    :param str unit: The unit of the value ('B', 'KB', 'MB', 'GB', 'TB').
    :param bool round_result: rounded result
    :return: The converted value and its unit.
    """

    if not value:
        return '0.00 MB'

    value = float(value)

    if value == 0:
        return "0.00 MB"

    units = ['B', 'KB', 'MB', 'GB', 'TB']
    if unit not in units:
        return "Invalid unit. Must be one of 'B', 'KB', 'MB', 'GB', 'TB'"

    unit_index = units.index(unit)

    while value > 999 and unit_index < len(units) - 1:
        value /= 1024
        unit_index += 1

    while value < 1 and unit_index > 0:
        value *= 1024
        unit_index -= 1

    if round_result:
        value = round(value / 5) * 5
    return f"{value:.2f} {units[unit_index]}"


def sum_sizes(size1, size2):
    """Sums two data storage values.

    :param str size1: The first value and its unit.
    :param str size2: The second value and its unit.
    :return: The sum of the values and its unit.
    """
    if size1 == '0.00 MB':
        return size2
    if size2 == '0.00 MB':
        return size1

    value1, unit1 = float(size1.split()[0]), size1.split()[1]
    value2, unit2 = float(size2.split()[0]), size2.split()[1]

    units = ['B', 'KB', 'MB', 'GB', 'TB']
    if unit1 not in units or unit2 not in units:
        return "Invalid unit in one of the sizes. Unit must be one of 'B', 'KB', 'MB', 'GB', 'TB'"

    value1 *= (1024 ** units.index(unit1))
    value2 *= (1024 ** units.index(unit2))

    total_value = value1 + value2

    total_unit = 'B'
    while total_value > 999 and total_unit != 'TB':
        total_value /= 1024
        total_unit = units[units.index(total_unit) + 1]

    return f"{total_value:.2f} {total_unit}"


def blank():
    """Returns a blank FormattedItem."""
    return FormattedItem(None, '-')


def listing(items, separator=','):
    """Given an iterable return a FormattedItem which display the list of items

        :param items: An iterable that outputs strings
        :param string separator: the separator to use
    """
    return SequentialOutput(separator, items)


def active_txn(item):
    """Returns a FormattedItem describing the active transaction on a object.

        If no active transaction is running, returns a blank FormattedItem.

        :param item: An object capable of having an active transaction
    """
    return transaction_status(utils.lookup(item, 'activeTransaction'))


def transaction_status(transaction):
    """Returns a FormattedItem describing the given transaction.

        :param item: An object capable of having an active transaction
    """
    if not transaction or not transaction.get('transactionStatus'):
        return blank()

    return FormattedItem(
        transaction['transactionStatus'].get('name'),
        transaction['transactionStatus'].get('friendlyName'))


def tags(tag_references):
    """Returns a formatted list of tags."""
    if not tag_references:
        return blank()

    tag_row = []
    for tag_detail in tag_references:
        tag = utils.lookup(tag_detail, 'tag', 'name')
        if tag is not None:
            tag_row.append(tag)

    return listing(tag_row, separator=', ')


def confirm(prompt_str, default=False):
    """Show a confirmation prompt to a command-line user.

    :param string prompt_str: prompt to give to the user
    :param bool default: Default value to True or False
    """
    if default:
        default_str = 'y'
        prompt = '%s [Y/n]' % prompt_str
    else:
        default_str = 'n'
        prompt = '%s [y/N]' % prompt_str

    try:
        ans = click.prompt(prompt, default=default_str, show_default=False)
    except click.exceptions.Abort:
        return False
    if ans.lower() in ('y', 'yes', 'yeah', 'yup', 'yolo'):
        return True

    return False


def no_going_back(confirmation):
    """Show a confirmation to a user.

    :param confirmation str: the string the user has to enter in order to confirm their action.
    """
    if not confirmation:
        confirmation = 'yes'

    prompt = f"This action cannot be undone! Type '{confirmation}' or press Enter to abort"

    try:
        ans = click.prompt(prompt, default='', show_default=False)
    except click.exceptions.Abort:
        return False
    if ans.lower() == str(confirmation).lower():
        return True

    return False


class SequentialOutput(list):
    """SequentialOutput is used for outputting sequential items.

    The purpose is to de-couple the separator from the output itself.

    :param separator str: string to use as a default separator
    """

    def __init__(self, separator=os.linesep, *args, **kwargs):
        self.separator = separator
        super().__init__(*args, **kwargs)

    def to_python(self):
        """returns itself, since it itself is a list."""
        return self

    def __str__(self):
        return self.separator.join(str(x) for x in self)


class CLIJSONEncoder(json.JSONEncoder):
    """A JSON encoder which is able to use a .to_python() method on objects."""

    def default(self, o):
        """Encode object if it implements to_python()."""
        if hasattr(o, 'to_python'):
            return o.to_python()
        return super().default(o)


class Table(object):
    """A Table structure used for output.

    :param list columns: a list of column names
    """

    def __init__(self, columns, title=None, align=None):
        duplicated_cols = [col for col, count
                           in collections.Counter(columns).items()
                           if count > 1]
        if len(duplicated_cols) > 0:
            raise exceptions.CLIAbort("Duplicated columns are not allowed: %s"
                                      % ','.join(duplicated_cols))

        self.columns = columns
        self.rows = []
        self.align = align or {}
        self.sortby = None
        self.title = title
        # Used to print a message if the table is empty
        self.empty_message = "-"

    def __bool__(self):
        """Useful for seeing if the table has any rows"""
        return len(self.rows) > 0

    def __str__(self):
        """A Table should only be cast to a string if its empty"""
        return self.empty_message

    def set_empty_message(self, message):
        """Sets the empty message for this table for env.fout

        Set this message if you want to print a message instead of a table to the user
        but still want the json output to print an empty list `[]`

        :param message str: Message to print if the table has no rows
        """
        self.empty_message = message

    def add_row(self, row):
        """Add a row to the table.

        :param list row: the row of string to be added
        """
        self.rows.append(row)

    def to_python(self):
        """Decode this Table object to standard Python types."""
        # Adding rows
        items = []
        for row in self.rows:
            formatted_row = [_format_python_value(v) for v in row]
            items.append(dict(zip(self.columns, formatted_row)))
        return items

    def prettytable(self, fmt='table', theme=None):
        """Returns a RICH table instance."""

        # Used to print a message instead of a bad looking empty table
        if not self and self.empty_message:
            return self.empty_message
        box_style = box.SQUARE
        if fmt == 'raw':
            box_style = None
        color_table = utils.table_color_theme(theme)
        table = rTable(title=self.title, box=box_style, header_style=color_table['header'])
        if self.sortby:
            try:
                # https://docs.python.org/3/howto/sorting.html#key-functions
                sort_index = self.columns.index(self.sortby)
                # All the values in `rows` are strings, so we need to cast to int for sorting purposes.
                self.rows.sort(key=lambda the_row: the_row[sort_index] if not the_row[sort_index].isdigit()
                               else int(the_row[sort_index]))
            except ValueError as ex:
                msg = "Column (%s) doesn't exist to sort by" % self.sortby
                raise exceptions.CLIAbort(msg) from ex

        for col in self.columns:
            justify = "center"
            style = None
            # This case aligns all columns in a table
            if isinstance(self.align, str):
                justify = self.align
            # This case alings a specific column
            elif isinstance(self.align, dict) and self.align.get(col, False):
                justify = self.align.get(col)
            # Backwards compatibility with PrettyTable style alignments
            if justify == 'r':
                justify = 'right'
            if justify == 'l':
                justify = 'left'
            # Special coloring for some columns
            if col in ('id', 'Id', 'ID'):
                style = color_table['id_columns']
            table.add_column(col, justify=justify, style=style)

        for row in self.rows:
            try:
                table.add_row(*row)
            # Generally you will see this if one of the columns in the row is a list or dict
            except NotRenderableError:
                forced_row = []
                for i in row:
                    forced_row.append(str(i))
                table.add_row(*forced_row)

        return table


class KeyValueTable(Table):
    """A table that is oriented towards key-value pairs."""

    def to_python(self):
        """Decode this KeyValueTable object to standard Python types."""
        mapping = {}
        for row in self.rows:
            mapping[row[0]] = _format_python_value(row[1])
        return mapping


class FormattedItem(object):
    """This is an object that can be displayed as a human readable and raw.

        :param original: raw (machine-readable) value
        :param string formatted: human-readable value
    """

    def __init__(self, original, formatted=None):
        self.original = original
        if formatted is not None:
            self.formatted = formatted
        else:
            self.formatted = self.original

    def to_python(self):
        """returns the original (raw) value."""
        return self.original

    def __str__(self):
        """returns the formatted value."""
        # If the original value is None, represent this as 'NULL'
        if self.original is None:
            return "NULL"

        try:
            return str(self.original)
        except UnicodeError:
            return 'invalid'

    def __repr__(self):
        return 'FormattedItem(%r, %r)' % (self.original, self.formatted)

    # Implement sorting methods.
    # NOTE(kmcdonald): functools.total_ordering could be used once support for
    # Python 2.6 is dropped
    def __eq__(self, other):
        return self.original == getattr(other, 'original', other)

    def __lt__(self, other):
        if self.original is None:
            return True

        other_val = getattr(other, 'original', other)
        if other_val is None:
            return False
        return self.original < other_val

    def __gt__(self, other):
        return not (self < other or self == other)

    def __le__(self, other):
        return self < other or self == other

    def __ge__(self, other):
        return self >= other


def _format_python_value(value):
    """If the value has to_python() defined then return that."""
    if hasattr(value, 'to_python'):
        return value.to_python()
    return value


def iter_to_table(value):
    """Convert raw API responses to response tables."""
    if isinstance(value, list):
        return _format_list(value)
    if isinstance(value, dict):
        return _format_dict(value)
    return value


def _format_dict(result):
    """Format dictionary responses into key-value table."""

    table = KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'

    for key, value in result.items():
        value = iter_to_table(value)
        table.add_row([key, value])

    return table


def _format_list(result):
    """Format list responses into a table."""

    if not result:
        return result
    table = Table(['value'])
    new_result = [item for item in result if item]
    if len(new_result) == 0:
        table.add_row(["-"])
        return table
    if isinstance(new_result[0], dict):
        return _format_list_objects(new_result)

    for item in new_result:
        table.add_row([iter_to_table(item)])
    return table


def _format_list_objects(result):
    """Format list of objects into a table."""

    all_keys = set()
    for item in result:
        if isinstance(item, dict):
            all_keys = all_keys.union(item.keys())

    all_keys = sorted(all_keys)
    table = Table(all_keys)

    for item in result:
        if not item:
            continue
        values = []
        for key in all_keys:
            value = iter_to_table(item.get(key))
            values.append(value)

        table.add_row(values)

    return table


def csv_output_format(data, delimiter=','):
    """Formating a table to csv format and show it."""
    data = clean_table(data, delimiter)
    write_csv_format(sys.stdout, data, delimiter)
    return ''


def clean_table(data, delimiter):
    """Delete Null fields by '-' and fix nested table in table"""
    new_data_row = []
    for row in data.rows:
        new_value = []
        for value in row:
            if str(value) == 'NULL':
                value = '-'

            if str(type(value)) == "<class 'SoftLayer.CLI.formatting.Table'>":
                string_io = io.StringIO()
                write_csv_format(string_io, value, delimiter, quoting=csv.QUOTE_MINIMAL)

                nested_table_converted = string_io.getvalue()
                nested_table_converted = nested_table_converted.replace('\r', '').split('\n')
                nested_table_converted.pop()

                title_nested_table = new_value.pop()
                for item in nested_table_converted:
                    new_value.append(title_nested_table)
                    new_value.append(item)
                    new_data_row.append(new_value)
                    new_value = []
            else:
                new_value.append(value)

        if len(new_value) != 0:
            new_data_row.append(new_value)
    data.rows = new_data_row
    return data


def write_csv_format(support_output, data, delimiter, quoting=csv.QUOTE_NONNUMERIC):
    """Write csv format to supported output"""
    writer = csv.writer(support_output, delimiter=delimiter, quoting=quoting)
    writer.writerow(data.columns)
    writer.writerows(data.rows)



"""
    SoftLayer.transports.xmlrpc
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~
    XML-RPC Style transport library

    :license: MIT, see LICENSE for more details.
"""
import re
from string import Template
import xmlrpc.client

import requests

from SoftLayer import consts
from SoftLayer import exceptions

from .transport import _format_object_mask
from .transport import _proxies_dict
from .transport import get_session
from .transport import SoftLayerListResult


class XmlRpcTransport(object):
    """XML-RPC transport."""

    def __init__(self, endpoint_url=None, timeout=None, proxy=None, user_agent=None, verify=True):

        self.endpoint_url = (endpoint_url or consts.API_PUBLIC_ENDPOINT).rstrip('/')
        self.timeout = timeout or None
        self.proxy = proxy
        self.user_agent = user_agent or consts.USER_AGENT
        self.verify = verify
        self._client = None

    @property
    def client(self):
        """Returns client session object"""

        if self._client is None:
            self._client = get_session(self.user_agent)
        return self._client

    def __call__(self, request):
        """Makes a SoftLayer API call against the XML-RPC endpoint.

        :param request request: Request object
        """
        largs = list(request.args)
        headers = request.headers

        auth = None
        if request.transport_user:
            auth = requests.auth.HTTPBasicAuth(request.transport_user, request.transport_password)

        if request.identifier is not None:
            header_name = request.service + 'InitParameters'
            headers[header_name] = {'id': request.identifier}

        if request.mask is not None:
            if isinstance(request.mask, dict):
                mheader = '%sObjectMask' % request.service
            else:
                mheader = 'SoftLayer_ObjectMask'
                request.mask = _format_object_mask(request.mask)
            headers.update({mheader: {'mask': request.mask}})

        if request.filter is not None:
            headers['%sObjectFilter' % request.service] = request.filter

        if request.limit:
            headers['resultLimit'] = {
                'limit': request.limit,
                'offset': request.offset or 0,
            }

        largs.insert(0, {'headers': headers})
        request.transport_headers.setdefault('Content-Type', 'application/xml')
        request.transport_headers.setdefault('User-Agent', self.user_agent)

        request.url = '/'.join([self.endpoint_url, request.service])
        request.payload = xmlrpc.client.dumps(tuple(largs),
                                              methodname=request.method,
                                              allow_none=True,
                                              encoding="iso-8859-1")

        # Prefer the request setting, if it's not None
        if request.verify is None:
            request.verify = self.verify

        try:
            resp = self.client.request('POST', request.url,
                                       data=request.payload.encode(),
                                       auth=auth,
                                       headers=request.transport_headers,
                                       timeout=self.timeout,
                                       verify=request.verify,
                                       cert=request.cert,
                                       proxies=_proxies_dict(self.proxy))

            resp.raise_for_status()
            result = xmlrpc.client.loads(resp.content)[0][0]
            if isinstance(result, list):
                return SoftLayerListResult(result, int(resp.headers.get('softlayer-total-items', 0)))
            else:
                return result
        except xmlrpc.client.Fault as ex:
            # These exceptions are formed from the XML-RPC spec
            # http://xmlrpc-epi.sourceforge.net/specs/rfc.fault_codes.php
            error_mapping = {
                '-32700': exceptions.NotWellFormed,
                '-32701': exceptions.UnsupportedEncoding,
                '-32702': exceptions.InvalidCharacter,
                '-32600': exceptions.SpecViolation,
                '-32601': exceptions.MethodNotFound,
                '-32602': exceptions.InvalidMethodParameters,
                '-32603': exceptions.InternalError,
                '-32500': exceptions.ApplicationError,
                '-32400': exceptions.RemoteSystemError,
                '-32300': exceptions.TransportError,
            }
            _ex = error_mapping.get(ex.faultCode, exceptions.SoftLayerAPIError)
            raise _ex(ex.faultCode, ex.faultString) from ex
        except requests.HTTPError as ex:
            err_message = f"{str(ex)} :: {ex.response.content}"
            raise exceptions.TransportError(ex.response.status_code, err_message)
        except requests.RequestException as ex:
            raise exceptions.TransportError(0, str(ex))

    def print_reproduceable(self, request):
        """Prints out the minimal python code to reproduce a specific request

        The will also automatically replace the API key so its not accidently exposed.

        :param request request: Request object
        """
        output = Template('''============= testing.py =============
import requests
from requests.auth import HTTPBasicAuth
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from xml.etree import ElementTree
client = requests.Session()
client.headers.update({'Content-Type': 'application/json', 'User-Agent': 'softlayer-python/testing',})
retry = Retry(connect=3, backoff_factor=3)
adapter = HTTPAdapter(max_retries=retry)
client.mount('https://', adapter)
# This is only needed if you are using an cloud.ibm.com api key
#auth=HTTPBasicAuth('apikey', YOUR_CLOUD_API_KEY)
auth=None
url = '$url'
payload = $payload
transport_headers = $transport_headers
timeout = $timeout
verify = $verify
cert = $cert
proxy = $proxy
response = client.request('POST', url, data=payload, headers=transport_headers, timeout=timeout,
               verify=verify, cert=cert, proxies=proxy, auth=auth)
xml = ElementTree.fromstring(response.content)
ElementTree.dump(xml)
==========================''')

        safe_payload = re.sub(r'<string>[a-z0-9]{64}</string>', r'<string>API_KEY_GOES_HERE</string>', request.payload)
        safe_payload = re.sub(r'(\s+)', r' ', safe_payload)
        safe_payload = safe_payload.encode()
        substitutions = {"url": request.url, "payload": safe_payload, "transport_headers": request.transport_headers,
                         "timeout": self.timeout, "verify": request.verify, "cert": request.cert,
                         "proxy": _proxies_dict(self.proxy)}
        return output.substitute(substitutions)



"""
    SoftLayer.testing
    ~~~~~~~~~~~~~~~~~

    :license: MIT, see LICENSE for more details.
"""
# Disable pylint import error and too many methods error
# pylint: disable=invalid-name
import logging
import os.path
import unittest
from unittest import mock as mock

from click import testing

import SoftLayer
from SoftLayer.CLI import core
from SoftLayer.CLI import environment
from SoftLayer.testing import xmlrpc

FIXTURE_PATH = os.path.abspath(os.path.join(__file__, '..', '..', 'fixtures'))


class MockableTransport(object):
    """Transport which is able to mock out specific API calls."""

    def __init__(self, transport):
        self.calls = []
        self.mocked = {}
        self.transport = transport

    def __call__(self, call):
        self._record_call(call)

        key = _mock_key(call.service, call.method)
        if key in self.mocked:
            return self.mocked[key](call)

        # Fall back to another transport (usually with fixtures)
        return self.transport(call)

    def set_mock(self, service, method):
        """Create a mock and return the mock object for the specific API call.

        :param service: API service to mock
        :param method: API method to mock
        """

        _mock = mock.MagicMock()
        self.mocked[_mock_key(service, method)] = _mock
        return _mock

    def clear(self):
        """Clear out mocks and call history."""
        self.calls = []
        self.mocked = {}

    def _record_call(self, call):
        """Record and log the API call (for later assertions)."""
        self.calls.append(call)

        details = []
        for prop in ['identifier',
                     'args',
                     'mask',
                     'filter',
                     'limit',
                     'offset']:
            details.append('%s=%r' % (prop, getattr(call, prop)))

        logging.info('%s::%s called; %s', call.service, call.method, '; '.join(details))


def _mock_key(service, method):
    """Key to address a mock object in MockableTransport."""
    return '%s::%s' % (service, method)


class TestCase(unittest.TestCase):
    """Testcase class with PEP-8 compatible method names."""

    @classmethod
    def setUpClass(cls):
        """Stand up fixtured/mockable XML-RPC server."""
        cls.mocks = MockableTransport(SoftLayer.FixtureTransport())
        cls.server = xmlrpc.create_test_server(cls.mocks)
        host, port = cls.server.socket.getsockname()[:2]
        cls.endpoint_url = "http://%s:%s" % (host, port)

    @classmethod
    def tearDownClass(cls):
        """Clean up the http server."""
        cls.server.shutdown()

    def set_up(self):
        """Aliased from setUp."""

    def tear_down(self):
        """Aliased from tearDown."""

    def setUp(self):  # NOQA
        unittest.TestCase.setUp(self)

        self.mocks.clear()

        transport = SoftLayer.XmlRpcTransport(endpoint_url=self.endpoint_url)
        wrapped_transport = SoftLayer.TimingTransport(transport)

        self.client = SoftLayer.BaseClient(transport=wrapped_transport)

        self.env = environment.Environment()
        self.env.client = self.client
        self.set_up()
        self.maxDiff = None

    def tearDown(self):  # NOQA
        super().tearDown()
        self.tear_down()
        self.mocks.clear()

    def calls(self, service=None, method=None, **props):
        """Return all API calls made during the current test."""

        conditions = []
        if service is not None:
            conditions.append(lambda call: call.service == service)
        if method is not None:
            conditions.append(lambda call: call.method == method)
        if props:
            conditions.append(lambda call: call_has_props(call, props))

        return [call for call in self.mocks.calls
                if all(cond(call) for cond in conditions)]

    def assert_called_with(self, service, method, **props):
        """Used to assert that API calls were called with given properties.

        Props are properties of the given transport.Request object.
        """

        if self.calls(service, method, **props):
            return

        raise AssertionError('%s::%s was not called with given properties: %s' % (service, method, props))

    def assert_not_called_with(self, service, method, **props):
        """Used to assert that API calls were NOT called with given properties.

        Props are properties of the given transport.Request object.
        """

        if self.calls(service, method, **props):
            raise AssertionError('%s::%s was called with given properties: %s' % (service, method, props))

    def assert_no_fail(self, result):
        """Fail when a failing click result has an error"""
        if result.exception:
            print(result.exception)
            raise result.exception

        self.assertEqual(result.exit_code, 0)

    def set_mock(self, service, method):
        """Set and return mock on the current client."""
        return self.mocks.set_mock(service, method)

    def run_command(self, args=None, env=None, fixtures=True, fmt='json', stdin=None):
        """A helper that runs a SoftLayer CLI command.

        This returns a click.testing.Result object.
        """
        args = args or []

        if fixtures is True:
            args.insert(0, '--demo')
        args.insert(0, '--format=%s' % fmt)

        runner = testing.CliRunner()
        return runner.invoke(core.cli, args=args, input=stdin, obj=env or self.env)

    def assertRaises(self, exception, function_callable, *args, **kwds):  # pylint: disable=arguments-differ
        """Converts testtools.assertRaises to unittest.assertRaises calls.

        testtools==2.4.0 require unittest2, which breaks pytest>=5.4.1 on skipTest.
        But switching to just using unittest breaks assertRaises because the format is slightly different.
        This basically just reformats the call so I don't have to re-write a bunch of tests.
        """
        with super().assertRaises(exception) as cm:
            function_callable(*args, **kwds)
        return cm.exception


def call_has_props(call, props):
    """Check if a call has matching properties of a given props dictionary."""

    for prop, expected_value in props.items():
        actual_value = getattr(call, prop)
        if actual_value != expected_value:
            logging.critical(
                '%s::%s property mismatch, %s: expected=%r; actual=%r',
                call.service,
                call.method,
                prop,
                expected_value,
                actual_value)
            return False

    return True



"""Lists all active billing items on this account. See https://cloud.ibm.com/billing/billing-items"""
# :license: MIT, see LICENSE for more details.
import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers.account import AccountManager as AccountManager
from SoftLayer import utils


@click.command(cls=SLCommand)
@click.option('--create', '-c', help='The date the billing item was created.')
@click.option('--ordered', '-o', help='Name that ordered the item')
@click.option('--category', '-C', help='Category name')
@environment.pass_env
def cli(env, create, category, ordered):
    """Lists billing items with some other useful information.

    Similiar to https://cloud.ibm.com/billing/billing-items
    """

    manager = AccountManager(env.client)
    items = manager.get_account_billing_items(create, category)
    table = item_table(items, ordered)

    env.fout(table)


def item_table(items, ordered=None):
    """Formats a table for billing items"""
    table = formatting.Table([
        "Id",
        "Create Date",
        "Cost",
        "Category Code",
        "Ordered By",
        "Description",
        "Notes"
    ], title="Billing Items")
    table.align['Description'] = 'l'
    table.align['Category Code'] = 'l'
    for item in items:
        description = item.get('description')
        fqdn = f"{item.get('hostName', '')}.{item.get('domainName', '')}"
        if fqdn != ".":
            description = fqdn
        user = utils.lookup(item, 'orderItem', 'order', 'userRecord')
        ordered_by = "IBM"
        create_date = utils.clean_time(item.get('createDate'), in_format='%Y-%m-%d', out_format='%Y-%m-%d')
        if user:
            # ordered_by = "{} ({})".format(user.get('displayName'), utils.lookup(user, 'userStatus', 'name'))
            ordered_by = user.get('displayName')
        if ordered:
            if ordered != ordered_by:
                continue
        table.add_row([
            item.get('id'),
            create_date,
            item.get('nextInvoiceTotalRecurringAmount'),
            item.get('categoryCode'),
            ordered_by,
            utils.trim_to(description, 50),
            utils.trim_to(item.get('notes', 'None'), 40),
        ])
    return table



"""Convert a dependent duplicate volume to an independent volume."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import environment


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('volume_id')
@environment.pass_env
def cli(env, volume_id):
    """Convert a dependent duplicate volume to an independent volume."""
    block_manager = SoftLayer.BlockStorageManager(env.client)
    resp = block_manager.convert_dep_dupe(volume_id)

    click.echo(resp)



"""List user notifications"""
import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@environment.pass_env
def cli(env):
    """My Notifications."""

    mgr = SoftLayer.UserManager(env.client)

    all_notifications = mgr.get_all_notifications()

    env.fout(notification_table(all_notifications))


def notification_table(all_notifications):
    """Creates a table of available notifications"""

    table = formatting.Table(['Id', 'Name', 'Description', 'Enabled'])
    table.align['Id'] = 'l'
    table.align['Name'] = 'l'
    table.align['Description'] = 'l'
    table.align['Enabled'] = 'l'
    for notification in all_notifications:
        table.add_row([notification['id'],
                       notification['name'],
                       notification['description'],
                       notification['enabled']])
    return table



getObject = {
    'id': 100,
    'hostname': 'vs-test1',
    'domain': 'test.sftlyr.ws',
    'fullyQualifiedDomainName': 'vs-test1.test.sftlyr.ws',
    'status': {'keyName': 'ACTIVE', 'name': 'Active'},
    'billingItem': {
        'id': 6327,
        'nextInvoiceTotalRecurringAmount': 1.54,
        'children': [
            {
                'categoryCode': 'ram',
                'description': '1 GB',
                'nextInvoiceTotalRecurringAmount': 1
            },
            {
                'categoryCode': 'remote_management',
                'description': 'Reboot / Remote Console',
                'nextInvoiceTotalRecurringAmount': 1
            },
            {
                'categoryCode': 'port_speed',
                'description': '1 Gbps Public & Private Network Uplinks',
                'nextInvoiceTotalRecurringAmount': 1
            },
            {
                'categoryCode': 'public_port',
                'description': '1 Gbps Public Uplink',
                'nextInvoiceTotalRecurringAmount': 1
            },
            {
                'categoryCode': 'service_port',
                'description': '1 Gbps Private Uplink',
                'nextInvoiceTotalRecurringAmount': 1
            }
        ],
        "nextInvoiceChildren": [
            {
                "allowCancellationFlag": 1,
                "associatedBillingItemId": "738974962",
                "cancellationDate": "2022-11-03T23:00:01-06:00",
                "categoryCode": "os_usage",
                "createDate": "2022-10-04T17:00:08-06:00",
                "cycleStartDate": "2022-10-04T17:00:08-06:00",
                "description": "CentOS 8.x Usage is 743.9997 hours Used",
                "id": 999303609,
                "laborFee": "0",
                "laborFeeTaxRate": "0",
                "lastBillDate": None,
                "modifyDate": None,
                "nextBillDate": "2022-11-03T23:00:00-06:00",
                "oneTimeFee": "0",
                "oneTimeFeeTaxRate": "0",
                "orderItemId": None,
                "parentId": None,
                "recurringFee": "0",
                "recurringFeeTaxRate": "0",
                "recurringMonths": 1,
                "serviceProviderId": 1,
                "setupFee": "0",
                "setupFeeTaxRate": "0",
                "associatedBillingItem": None
            },
            {
                "allowCancellationFlag": 1,
                "associatedBillingItemId": "738974962",
                "cancellationDate": "2022-11-03T23:00:01-06:00",
                "categoryCode": "guest_core_usage",
                "createDate": "2022-10-04T17:00:09-06:00",
                "cycleStartDate": "2022-10-04T17:00:09-06:00",
                "description": "CPU Cores: a suspendable product.",
                "id": 999303621,
                "laborFee": "0",
                "laborFeeTaxRate": "0",
                "lastBillDate": None,
                "modifyDate": None,
                "nextBillDate": "2022-11-03T23:00:00-06:00",
                "oneTimeFee": "0",
                "oneTimeFeeTaxRate": "0",
                "orderItemId": None,
                "parentId": None,
                "recurringFee": "1.54",
                "recurringFeeTaxRate": "0",
                "recurringMonths": 1,
                "serviceProviderId": 1,
                "setupFee": "0",
                "setupFeeTaxRate": "0",
                "associatedBillingItem": None
            },
        ],
        'package': {
            "id": 835,
            "keyName": "PUBLIC_CLOUD_SERVER"
        },
        'orderItem': {
            'order': {
                'userRecord': {
                    'username': 'chechu',
                }
            },
            'preset': {'keyName': 'B1_8X16X100'}
        }
    },
    'datacenter': {'id': 50, 'name': 'TEST00', 'longName': 'test 00',
                   'description': 'Test Data Center'},
    'powerState': {'keyName': 'RUNNING', 'name': 'Running'},
    'maxCpu': 2,
    'maxMemory': 1024,
    'primaryIpAddress': '172.16.240.2',
    'globalIdentifier': '1a2b3c-1701',
    'primaryBackendIpAddress': '10.45.19.37',
    'primaryNetworkComponent': {'speed': 10, 'maxSpeed': 100},
    'hourlyBillingFlag': False,
    'createDate': '2013-08-01 15:23:45',
    'blockDevices': [{'device': 0, 'mountType': 'Disk', 'uuid': 1},
                     {'device': 1, 'mountType': 'Disk',
                      'diskImage': {'type': {'keyName': 'SWAP'}}},
                     {'device': 2, 'mountType': 'CD'},
                     {'device': 3, 'mountType': 'Disk', 'uuid': 3},
                     {'device': 4, 'mountType': 'Disk', 'uuid': 4,
                      'diskImage': {'metadataFlag': True}}],
    'notes': 'notes',
    'networkVlans': [{'networkSpace': 'PUBLIC',
                      'vlanNumber': 23,
                      'id': 1}],
    'dedicatedHost': {'id': 37401},
    'transientGuestFlag': False,
    'operatingSystem': {
        'passwords': [{'username': 'user', 'password': 'pass'}],
        'softwareLicense': {
            'softwareDescription': {'version': '12.04-64 Minimal for VSI',
                                    'name': 'Ubuntu', 'referenceCode': 'UBUNTU_8_64', }}
    },
    'softwareComponents': [{
        'passwords': [{'username': 'user', 'password': 'pass'}],
        'softwareLicense': {
            'softwareDescription': {'name': 'Ubuntu'}}
    }],
    'tagReferences': [{'tag': {'name': 'production'}}],
    'networkMonitors': [
        {
            'guestId': 116114480,
            'hostId': 116114480,
            'id': 17653845,
            'ipAddress': '52.116.23.73',
            'queryTypeId': 1,
            'responseActionId': 1,
            'status': 'ON',
            'waitCycles': 0,
            'queryType': {
                'description': 'Test ping to address',
                'id': 1,
                'monitorLevel': 0,
                'name': 'SERVICE PING'
            },
            'responseAction': {
                'actionDescription': 'Do Nothing',
                'id': 1,
                'level': 0
            }
        }
    ]
}
getCreateObjectOptions = {
    'flavors': [
        {
            'flavor': {
                'keyName': 'B1_1X2X25',
                'name': 'B1-1X2X25'
            },
            'template': {
                'supplementalCreateObjectOptions': {
                    'flavorKeyName': 'B1_1X2X25'
                }
            }
        },
        {
            'flavor': {
                'keyName': 'B1_1X2X25_TRANSIENT',
                'name': 'B1-1X2X25_TRANSIENT'
            },
            'template': {
                'supplementalCreateObjectOptions': {
                    'flavorKeyName': 'B1_1X2X25_TRANSIENT'
                },
                'transientGuestFlag': True
            }
        },
        {
            'flavor': {
                'keyName': 'B1_1X2X100',
                'name': 'B1-1X2X100'
            },
            'template': {
                'supplementalCreateObjectOptions': {
                    'flavorKeyName': 'B1_1X2X100'
                }
            }
        },
        {
            'flavor': {
                'keyName': 'BL1_1X2X100',
                'name': 'BL1_1X2X100'
            },
            'template': {
                'supplementalCreateObjectOptions': {
                    'flavorKeyName': 'BL1_1X2X100'
                }
            }
        },
        {
            'flavor': {
                'keyName': 'BL2_1X2X100',
                'name': 'BL2-1X2X100'
            },
            'template': {
                'supplementalCreateObjectOptions': {
                    'flavorKeyName': 'BL2_1X2X100'
                }
            }
        },
        {
            'flavor': {
                'keyName': 'C1_1X2X25',
                'name': 'C1-1X2X25'
            },
            'template': {
                'supplementalCreateObjectOptions': {
                    'flavorKeyName': 'C1_1X2X25'
                }
            }
        },
        {
            'flavor': {
                'keyName': 'M1_1X2X100',
                'name': 'M1_1X2X100'
            },
            'template': {
                'supplementalCreateObjectOptions': {
                    'flavorKeyName': 'M1_1X2X100'
                }
            }
        },
        {
            'flavor': {
                'keyName': 'AC1_1X2X100',
                'name': 'AC1_1X2X100'
            },
            'template': {
                'supplementalCreateObjectOptions': {
                    'flavorKeyName': 'AC1_1X2X100'
                }
            }
        },
        {
            'flavor': {
                'keyName': 'ACL1_1X2X100',
                'name': 'ACL1_1X2X100'
            },
            'template': {
                'supplementalCreateObjectOptions': {
                    'flavorKeyName': 'ACL1_1X2X100'
                }
            }
        },
    ],
    'processors': [
        {
            'itemPrice': {
                'item': {'description': '1 x 2.0 GHz Core'},
                'hourlyRecurringFee': '.07',
                'recurringFee': '29'
            },
            'template': {'startCpus': 1}
        },
        {
            'itemPrice': {
                'item': {'description': '2 x 2.0 GHz Cores'},
                'hourlyRecurringFee': '.14',
                'recurringFee': '78'
            },
            'template': {'startCpus': 2}
        },
        {
            'itemPrice': {
                'item': {'description': '3 x 2.0 GHz Cores'},
                'hourlyRecurringFee': '.205',
                'recurringFee': '123.5'
            },
            'template': {'startCpus': 3}
        },
        {
            'itemPrice': {
                'item': {'description': '4 x 2.0 GHz Cores'},
                'hourlyRecurringFee': '.265',
                'recurringFee': '165.5'
            },
            'template': {'startCpus': 4}
        },
        {
            'itemPrice': {
                'hourlyRecurringFee': '.209',
                'recurringFee': '139',
                'dedicatedHostInstanceFlag': False,
                'item': {
                    'description': '1 x 2.0 GHz Cores (Dedicated)'
                }
            },
            'template': {
                'dedicatedAccountHostOnlyFlag': True,
                'startCpus': 1
            }
        },
        {
            'itemPrice': {
                'hourlyRecurringFee': '0',
                'recurringFee': '0',
                'dedicatedHostInstanceFlag': True,
                'item': {
                    'description': '56 x 2.0 GHz Cores (Dedicated Host)'
                }
            },
            'template': {
                'startCpus': 56,
                'dedicatedHost': {
                    'id': None
                }
            }
        },
        {
            'itemPrice': {
                'hourlyRecurringFee': '0',
                'recurringFee': '0',
                'dedicatedHostInstanceFlag': True,
                'item': {
                    'description': '4 x 2.0 GHz Cores (Dedicated Host)'
                }
            },
            'template': {
                'startCpus': 4,
                'dedicatedHost': {
                    'id': None
                }
            }
        },
    ],
    'memory': [
        {
            "description": "1 GB ",
            "attributes": [],
            "itemCategory": {
                "categoryCode": "ram",
                "id": 4
            },
            'itemPrice': {
                'item': {'description': '1 GB'},
                'hourlyRecurringFee': '.03',
                'recurringFee': '21'
            },
            'template': {'maxMemory': 1024}
        },
        {
            "description": "2 GB ",
            "attributes": [],
            "itemCategory": {
                "categoryCode": "ram",
                "id": 5
            },
            'itemPrice':
                {
                    'item': {'description': '2 GB'},
                    'hourlyRecurringFee': '.06',
                    'recurringFee': '42'
                },
            'template': {'maxMemory': 2048}
        },
        {
            "description": "3 GB",
            "attributes": [],
            "itemCategory": {
                "categoryCode": "ram",
                "id": 65
            },
            'itemPrice': {
                'item': {'description': '3 GB'},
                'hourlyRecurringFee': '.085',
                'recurringFee': '59.5'},
            'template': {'maxMemory': 3072}
        },
        {
            "description": "4 GB",
            "attributes": [],
            "itemCategory": {
                "categoryCode": "ram",
                "id": 35
            },
            'itemPrice': {
                'item': {'description': '4 GB'},
                'hourlyRecurringFee': '.11',
                'recurringFee': '77'
            },
            'template': {'maxMemory': 4096}
        },
        {
            "description": "64 GB (Dedicated Host)",
            "attributes": [],
            "itemCategory": {
                "categoryCode": "ram",
                "id": 3
            },
            'itemPrice': {
                'hourlyRecurringFee': '0',
                'recurringFee': '0',
                'dedicatedHostInstanceFlag': True,
                'item': {
                    'description': '64 GB (Dedicated Host)'
                }
            },
            'template': {
                'maxMemory': 65536
            }
        },
        {
            "description": "8 GB (Dedicated Host)",
            "attributes": [],
            "itemCategory": {
                "categoryCode": "ram",
                "id": 36
            },
            'itemPrice': {
                'hourlyRecurringFee': '0',
                'recurringFee': '0',
                'dedicatedHostInstanceFlag': True,
                'item': {
                    'description': '8 GB (Dedicated Host)'
                }
            },
            'template': {
                'maxMemory': 8192
            }
        },
    ],
    'blockDevices': [
        {
            'itemPrice': {
                'item': {'description': '25 GB (LOCAL)'},
                'hourlyRecurringFee': '0',
                'recurringFee': '0'},
            'template': {
                'blockDevices': [
                    {'device': '0', 'diskImage': {'capacity': 25}}
                ],
                'localDiskFlag': True
            }
        },
        {
            'itemPrice': {
                'item': {'description': '100 GB (LOCAL)'},
                'hourlyRecurringFee': '.01',
                'recurringFee': '7'
            },
            'template': {
                'blockDevices': [
                    {'device': '0', 'diskImage': {'capacity': 100}}
                ],
                'localDiskFlag': True
            }
        },
    ],
    'operatingSystems': [
        {
            'itemPrice': {
                'item': {
                    'description': 'CentOS 6.0 - Minimal Install (64 bit)'
                },
                'hourlyRecurringFee': '0',
                'recurringFee': '0'
            },
            'template': {
                'operatingSystemReferenceCode': 'CENTOS_6_64'
            }
        },
        {
            'itemPrice': {
                'item': {
                    'description': 'Debian GNU/Linux 7.0 Wheezy/Stable -'
                                   ' Minimal Install (64 bit)'
                },
                'hourlyRecurringFee': '0',
                'recurringFee': '0'
            },
            'template': {
                'operatingSystemReferenceCode': 'DEBIAN_7_64'
            }
        },
        {
            'itemPrice': {
                'item': {
                    'description': 'Ubuntu Linux 12.04 LTS Precise'
                                   ' Pangolin - Minimal Install (64 bit)'
                },
                'hourlyRecurringFee': '0',
                'recurringFee': '0'
            },
            'template': {
                'operatingSystemReferenceCode': 'UBUNTU_12_64'
            }
        },
    ],
    'networkComponents': [
        {
            'itemPrice': {
                'item': {
                    'description': '10 Mbps Public & Private Networks'
                },
                'hourlyRecurringFee': '0',
                'recurringFee': '0'},
            'template': {
                'networkComponents': [{'maxSpeed': 10}]
            }
        },
        {
            'itemPrice': {
                'item': {'description': '100 Mbps Private Network'},
                'hourlyRecurringFee': '0',
                'recurringFee': '0'},
            'template': {
                'networkComponents': [{'maxSpeed': 100}]
            }
        },
        {
            'itemPrice': {
                'item': {'description': '1 Gbps Private Network'},
                'hourlyRecurringFee': '.02',
                'recurringFee': '10'
            },
            'template': {
                'networkComponents': [{'maxSpeed': 1000}]
            }
        },
        {
            'itemPrice': {
                'hourlyRecurringFee': '0',
                'recurringFee': '0',
                'dedicatedHostInstanceFlag': True,
                'item': {
                    'description': '1 Gbps Public & Private Network Uplinks (Dedicated Host)'
                }
            },
            'template': {
                'networkComponents': [
                    {
                        'maxSpeed': 1000
                    }
                ],
                'privateNetworkOnlyFlag': False
            }
        },
    ],
    'datacenters': [
        {'template': {'datacenter': {'name': 'ams01'}}},
        {'template': {'datacenter': {'name': 'dal05'}}},
    ],
    'guest_disk': [{
        "description": "25 GB (SAN)",
        "attributes": [
            {
                "id": 197,
                "attributeTypeKeyName": "SAN_DISK"
            }
        ],
        "itemCategory": {
            "categoryCode": "guest_disk0",
            "id": 81
        }}, {
        "description": "250 GB (SAN)",
        "attributes": [
            {
                "id": 198,
                "attributeTypeKeyName": "SAN_DISK"
            }],
        "itemCategory": {
            "categoryCode": "guest_disk0",
            "id": 89
        }}],
    'guest_core': [{
        "description": "4 x 2.0 GHz or higher Cores (Dedicated)",
        "attributes": [],
        "itemCategory": {
            "categoryCode": "guest_core",
            "id": 80
        }},
        {
            "description": "8 x 2.0 GHz or higher Cores",
            "attributes": [],
            "itemCategory": {
                "categoryCode": "guest_core",
                "id": 90
            }}]
}

getReverseDomainRecords = [{
    'networkAddress': '12.34.56.78',
    'name': '12.34.56.78.in-addr.arpa',
    'resourceRecords': [{'data': 'test.softlayer.com.', 'id': 987654}],
    'updateDate': '2013-09-11T14:36:57-07:00',
    'serial': 1234665663,
    'id': 123456,
}]

editObject = True
deleteObject = True
setPrivateNetworkInterfaceSpeed = True
setPublicNetworkInterfaceSpeed = True
createObject = getObject
createObjects = [getObject]
generateOrderTemplate = {
    "imageTemplateId": None,
    "location": "1854895",
    "packageId": 835,
    "presetId": 405,
    "prices": [
        {
            "hourlyRecurringFee": "0",
            "id": 45466,
            "recurringFee": "0",
            "item": {
                "description": "CentOS 7.x - Minimal Install (64 bit)"
            }
        },
        {
            "hourlyRecurringFee": "0",
            "id": 2202,
            "recurringFee": "0",
            "item": {
                "description": "25 GB (SAN)"
            }
        },
        {
            "hourlyRecurringFee": "0",
            "id": 905,
            "recurringFee": "0",
            "item": {
                "description": "Reboot / Remote Console"
            }
        },
        {
            "hourlyRecurringFee": ".02",
            "id": 899,
            "recurringFee": "10",
            "item": {
                "description": "1 Gbps Private Network Uplink"
            }
        },
        {
            "hourlyRecurringFee": "0",
            "id": 1800,
            "item": {
                "description": "0 GB Bandwidth Allotment"
            }
        },
        {
            "hourlyRecurringFee": "0",
            "id": 21,
            "recurringFee": "0",
            "item": {
                "description": "1 IP Address"
            }
        },
        {
            "hourlyRecurringFee": "0",
            "id": 55,
            "recurringFee": "0",
            "item": {
                "description": "Host Ping"
            }
        },
        {
            "hourlyRecurringFee": "0",
            "id": 57,
            "recurringFee": "0",
            "item": {
                "description": "Email and Ticket"
            }
        },
        {
            "hourlyRecurringFee": "0",
            "id": 58,
            "recurringFee": "0",
            "item": {
                "description": "Automated Notification"
            }
        },
        {
            "hourlyRecurringFee": "0",
            "id": 420,
            "recurringFee": "0",
            "item": {
                "description": "Unlimited SSL VPN Users & 1 PPTP VPN User per account"
            }
        },
    ],
    "quantity": 1,
    "sourceVirtualGuestId": None,
    "sshKeys": [],
    "useHourlyPricing": True,
    "virtualGuests": [
        {
            "domain": "test.local",
            "hostname": "test"
        }
    ],
    "complexType": "SoftLayer_Container_Product_Order_Virtual_Guest"
}

setUserMetadata = ['meta']
reloadOperatingSystem = 'OK'
setTags = True
createArchiveTemplate = {
    'createDate': '2018-12-10T17:29:18-06:00',
    'elapsedSeconds': 0,
    'guestId': 12345678,
    'hardwareId': None,
    'id': 12345,
    'modifyDate': '2018-12-10T17:29:18-06:00',
    'statusChangeDate': '2018-12-10T17:29:18-06:00',
    'globalIdentifier': '495c47c2-80af-4fef-87a9-97faf132345678'
}

executeRescueLayer = True

getUpgradeItemPrices = [
    {
        'id': 1007,
        'categories': [{'id': 80,
                        'name': 'Computing Instance',
                        'categoryCode': 'guest_core'}],
        'item': {
            'capacity': '4',
            'units': 'PRIVATE_CORE',
            'description': 'Computing Instance (Dedicated)',
        }
    },
    {
        'id': 1144,
        'locationGroupId': None,
        'categories': [{'id': 80,
                        'name': 'Computing Instance',
                        'categoryCode': 'guest_core'}],
        'item': {
            'capacity': '4',
            'units': 'CORE',
            'description': 'Computing Instance',
        }
    },
    {
        'id': 332211,
        'locationGroupId': 1,
        'categories': [{'id': 80,
                        'name': 'Computing Instance',
                        'categoryCode': 'guest_core'}],
        'item': {
            'capacity': '4',
            'units': 'CORE',
            'description': 'Computing Instance',
        }
    },
    {
        'id': 1122,
        'categories': [{'id': 26,
                        'name': 'Uplink Port Speeds',
                        'categoryCode': 'port_speed'}],
        'item': {
            'capacity': '1000',
            'description': 'Public & Private Networks',
        }
    },
    {
        'id': 1144,
        'categories': [{'id': 26,
                        'name': 'Uplink Port Speeds',
                        'categoryCode': 'port_speed'}],
        'item': {
            'capacity': '1000',
            'description': 'Private Networks',
        }
    },
    {
        'id': 1133,
        'categories': [{'id': 3,
                        'name': 'RAM',
                        'categoryCode': 'ram'}],
        'item': {
            'capacity': '2',
            'description': 'RAM',
        }
    }, {
        "id": 2255,
        "categories": [
            {
                "categoryCode": "guest_disk1",
                "id": 82,
                "name": "Second Disk"
            },
            {
                "categoryCode": "guest_disk2",
                "id": 92,
                "name": "Third Disk"
            },
            {
                "categoryCode": "guest_disk3",
                "id": 93,
                "name": "Fourth Disk"
            },
            {
                "categoryCode": "guest_disk4",
                "id": 116,
                "name": "Fifth Disk"
            }
        ],
        "item": {
            "capacity": "10",
            "description": "10 GB (SAN)"
        }
    }
]

DEDICATED_GET_UPGRADE_ITEM_PRICES = [
    {
        'id': 115566,
        'categories': [{'id': 80,
                        'name': 'Computing Instance',
                        'categoryCode': 'guest_core'}],
        'item': {
            'capacity': '4',
            'units': 'DEDICATED_CORE',
            'description': 'Computing Instance (Dedicated Host)',
        }
    },
]

getMetricTrackingObjectId = 1000

getBandwidthAllotmentDetail = {
    'allocationId': 25465663,
    'bandwidthAllotmentId': 138442,
    'effectiveDate': '2019-04-03T23:00:00-06:00',
    'endEffectiveDate': None,
    'id': 25888247,
    'serviceProviderId': 1,
    'allocation': {
        'amount': '250'
    }
}

getBillingCycleBandwidthUsage = [
    {
        'amountIn': '.448',
        'amountOut': '.52157',
        'type': {
            'alias': 'PUBLIC_SERVER_BW'
        }
    },
    {
        'amountIn': '.03842',
        'amountOut': '.01822',
        'type': {
            'alias': 'PRIVATE_SERVER_BW'
        }
    }
]

getAttachedNetworkStorages = [
    {
        "accountId": 11111,
        "capacityGb": 20,
        "createDate": "2018-04-05T05:15:49-06:00",
        "id": 22222,
        "nasType": "NAS",
        "serviceProviderId": 1,
        "storageTypeId": "13",
        "username": "SL02SEV311111_11",
        "allowedVirtualGuests": [
            {
                "id": 12345,
                "datacenter": {
                    "id": 449506,
                    "longName": "Frankfurt 2",
                    "name": "fra02",
                    "statusId": 2
                }
            }
        ],
        "serviceResourceBackendIpAddress": "fsn-fra0201a-fz.service.softlayer.com",
        "serviceResourceName": "Storage Type 02 File Aggregate stfm-fra0201a"
    },
    {
        "accountId": 11111,
        "capacityGb": 12000,
        "createDate": "2018-01-28T04:57:30-06:00",
        "id": 3777111,
        "nasType": "ISCSI",
        "notes": "BlockStorage12T",
        "password": "",
        "serviceProviderId": 1,
        "storageTypeId": "7",
        "username": "SL02SEL32222-9",
        "allowedVirtualGuests": [
            {
                "id": 629222,
                "datacenter": {
                    "id": 449506,
                    "longName": "Frankfurt 2",
                    "name": "fra02",
                    "statusId": 2
                }
            }
        ],
        "serviceResourceBackendIpAddress": "10.31.95.152",
        "serviceResourceName": "Storage Type 02 Block Aggregate stbm-fra0201a"
    }
]

getAllowedHost = {
    "accountId": 11111,
    "credentialId": 22222,
    "id": 33333,
    "name": "iqn.2020-03.com.ibm:sl02su11111-v62941551",
    "resourceTableId": 6291111,
    "resourceTableName": "VIRTUAL_GUEST",
    "credential": {
        "accountId": "11111",
        "createDate": "2020-03-20T13:35:47-06:00",
        "id": 44444,
        "nasCredentialTypeId": 2,
        "password": "SjFDCpHrjskfj",
        "username": "SL02SU11111-V62941551"
    }
}

getBlockDevices = [
    {
        "createDate": "2018-10-06T04:27:35-06:00",
        "device": "0",
        "id": 11111,
        "mountType": "Disk",
        "diskImage": {
            "capacity": 100,
            "description": "adns.vmware.com",
            "id": 72222,
            "name": "adns.vmware.com",
            "units": "GB",
        }
    },
    {
        "device": "1",
        "id": 22222,
        "mountType": "Disk",
        "statusId": 1,
        "diskImage": {
            "capacity": 2,
            "description": "6211111-SWAP",
            "id": 33333,
            "name": "6211111-SWAP",
            "units": "GB",
        }
    }
]

migrate = True
migrateDedicatedHost = True
allowAccessToNetworkStorageList = True

attachDiskImage = {
    "createDate": "2021-03-22T13:15:31-06:00",
    "id": 1234567
}

getBrowserConsoleAccessLogs = [
    {
        "createDate": "2022-03-07T05:32:23-06:00",
        "eventType": "CONNECTED",
        "id": 509636,
        "message": "User connected",
        "sourceIp": "44.200.9.0",
        "sourcePort": 47097,
        "username": "testuser@ibm.com@ibm.com"
    },
    {
        "createDate": "2022-03-07T07:58:27-06:00",
        "eventType": "CONNECTED",
        "id": 509024,
        "message": "User connected",
        "sourceIp": "44.200.9.0",
        "sourcePort": 41185,
        "username": "testuser@ibm.com"
    }]



"""
    SoftLayer.CLI.helpers
    ~~~~~~~~~~~~~~~~~~~~~
    Helpers to be used in CLI modules in SoftLayer.CLI.modules.*

    :license: MIT, see LICENSE for more details.
"""

import click

from SoftLayer.CLI import exceptions


def multi_option(*param_decls, **attrs):
    """modify help text and indicate option is permitted multiple times

    :param param_decls:
    :param attrs:
    :return:

    """
    attrhelp = attrs.get('help', None)
    if attrhelp is not None:
        newhelp = attrhelp + " (multiple occurrence permitted)"
        attrs['help'] = newhelp
    attrs['multiple'] = True
    return click.option(*param_decls, **attrs)


def resolve_id(resolver, identifier, name='object'):
    """Resolves a single id using a resolver function.

    :param resolver: function that resolves ids. Should return None or a list of ids.
    :param string identifier: a string identifier used to resolve ids
    :param string name: the object type, to be used in error messages

    """
    try:
        return int(identifier)
    except ValueError:
        pass  # It was worth a shot

    ids = resolver(identifier)

    if len(ids) == 0:
        raise exceptions.CLIAbort("Error: Unable to find %s '%s'" % (name, identifier))

    if len(ids) > 1:
        raise exceptions.CLIAbort(
            "Error: Multiple %s found for '%s': %s" %
            (name, identifier, ', '.join([str(_id) for _id in ids])))

    return ids[0]



"""Power commands."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--force', default=False, is_flag=True, help="Force modify")
@environment.pass_env
def power_off(env, identifier, force):
    """Power off an active server."""

    mgr = SoftLayer.HardwareManager(env.client)
    hw_id = helpers.resolve_id(mgr.resolve_ids, identifier, 'hardware')
    if not force:
        if not (env.skip_confirmations or
                formatting.confirm('This will power off the server with id %s '
                                   'Continue?' % hw_id)):
            raise exceptions.CLIAbort('Aborted.')

    env.client['Hardware_Server'].powerOff(id=hw_id)


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--hard/--soft',
              default=None,
              help="Perform a hard or soft reboot")
@environment.pass_env
def reboot(env, identifier, hard):
    """Reboot an active server."""

    hardware_server = env.client['Hardware_Server']
    mgr = SoftLayer.HardwareManager(env.client)
    hw_id = helpers.resolve_id(mgr.resolve_ids, identifier, 'hardware')
    if not (env.skip_confirmations or
            formatting.confirm('This will power off the server with id %s. '
                               'Continue?' % hw_id)):
        raise exceptions.CLIAbort('Aborted.')

    if hard is True:
        hardware_server.rebootHard(id=hw_id)
    elif hard is False:
        hardware_server.rebootSoft(id=hw_id)
    else:
        hardware_server.rebootDefault(id=hw_id)


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--force', default=False, is_flag=True, help="Force modify")
@environment.pass_env
def power_on(env, identifier, force):
    """Power on a server."""

    mgr = SoftLayer.HardwareManager(env.client)
    hw_id = helpers.resolve_id(mgr.resolve_ids, identifier, 'hardware')

    if not force:
        if not (env.skip_confirmations or
                formatting.confirm('This will power off the server with id %s. '
                                   'Continue?' % hw_id)):
            raise exceptions.CLIAbort('Aborted.')

    env.client['Hardware_Server'].powerOn(id=hw_id)


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def power_cycle(env, identifier):
    """Power cycle a server."""

    mgr = SoftLayer.HardwareManager(env.client)
    hw_id = helpers.resolve_id(mgr.resolve_ids, identifier, 'hardware')

    if not (env.skip_confirmations or
            formatting.confirm('This will power off the server with id %s. '
                               'Continue?' % hw_id)):
        raise exceptions.CLIAbort('Aborted.')

    env.client['Hardware_Server'].powerCycle(id=hw_id)


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def rescue(env, identifier):
    """Reboot server into a rescue image."""

    mgr = SoftLayer.HardwareManager(env.client)
    hw_id = helpers.resolve_id(mgr.resolve_ids, identifier, 'hardware')

    if not (env.skip_confirmations or
            formatting.confirm("This action will reboot this server. Continue?")):

        raise exceptions.CLIAbort('Aborted')

    env.client['Hardware_Server'].bootToRescueLayer(id=hw_id)



"""
    SoftLayer.auth
    ~~~~~~~~~~~~~~
    Module with the supported auth mechanisms for the SoftLayer API

    :license: MIT, see LICENSE for more details.
"""
import os

__all__ = [
    'BasicAuthentication',
    'TokenAuthentication',
    'BasicHTTPAuthentication',
    'AuthenticationBase',
    'X509Authentication',
    'EmployeeAuthentication'
]


class AuthenticationBase(object):
    """A base authentication class intended to be overridden."""

    def get_request(self, request):
        """Receives request options and returns request options.

            :param options dict: dictionary of request options

        """
        return request

    def get_headers(self):
        """Return a dictionary of headers to be inserted for authentication.

        .. deprecated:: 3.3.0
           Use :func:`get_options` instead.
        """
        return {}


class TokenAuthentication(AuthenticationBase):
    """Token-based authentication class.

        :param user_id int: a user's id
        :param auth_token str: a user's auth token, attained through
                               User_Customer::getPortalLoginToken
    """

    def __init__(self, user_id, auth_token):
        self.user_id = user_id
        self.auth_token = auth_token

    def get_request(self, request):
        """Sets token-based auth headers."""
        request.headers['authenticate'] = {
            'complexType': 'PortalLoginToken',
            'userId': self.user_id,
            'authToken': self.auth_token,
        }
        return request

    def __repr__(self):
        return "TokenAuthentication(%r)" % self.user_id


class BasicAuthentication(AuthenticationBase):
    """Token-based authentication class.

        :param username str: a user's username
        :param api_key str: a user's API key
    """

    def __init__(self, username, api_key):
        self.username = username
        self.api_key = api_key

    def get_request(self, request):
        """Sets token-based auth headers."""

        # See https://cloud.ibm.com/docs/iam?topic=iam-iamapikeysforservices for why this is the way it is
        if self.username == 'apikey':
            request.transport_user = self.username
            request.transport_password = self.api_key
        else:
            request.headers['authenticate'] = {
                'username': self.username,
                'apiKey': self.api_key,
            }

        return request

    def __repr__(self):
        return f"BasicAuthentication(username={self.username})"


class BasicHTTPAuthentication(AuthenticationBase):
    """Token-based authentication class.

        :param username str: a user's username
        :param api_key str: a user's API key
    """

    def __init__(self, username, api_key):
        self.username = username
        self.api_key = api_key

    def get_request(self, request):
        """Sets token-based auth headers."""
        request.transport_user = self.username
        request.transport_password = self.api_key
        return request

    def __repr__(self):
        return f"BasicHTTPAuthentication(username={self.username}"


class BearerAuthentication(AuthenticationBase):
    """Bearer Token authentication class.

        :param username str: a user's username, not really needed but all the others use it.
        :param api_key str: a user's IAM Token
    """

    def __init__(self, username, token, r_token=None):
        """For using IBM IAM authentication

        :param username str: Not really needed, will be set to their current username though for logging
        :param token str: the IAM Token
        :param r_token str: The refresh Token, optional
        """
        self.username = username
        self.api_key = token
        self.r_token = r_token

    def get_request(self, request):
        """Sets token-based auth headers."""
        request.transport_headers['Authorization'] = f'Bearer {self.api_key}'
        request.transport_user = self.username
        return request

    def __repr__(self):
        return f"BearerAuthentication(username={self.username}, token={self.api_key})"


class X509Authentication(AuthenticationBase):
    """X509Authentication authentication class.

        :param certificate str:  Path to a users SSL certificate for authentication
        :param CA Cert str: Path to the Servers signed certificate.
    """

    def __init__(self, cert, ca_cert):
        self.cert = os.path.expanduser(cert)
        self.ca_cert = ca_cert

    def get_request(self, request):
        """Sets token-based auth headers."""
        request.cert = self.cert
        request.verify = self.ca_cert
        return request

    def __repr__(self):
        return f"X509Authentication(cert={self.cert}, ca_cert={self.ca_cert})"


class EmployeeAuthentication(AuthenticationBase):
    """Token-based authentication class.

        :param username str: a user's username
        :param user_hash str: a user's Authentication hash
    """
    def __init__(self, user_id, user_hash):
        self.user_id = user_id
        self.hash = user_hash

    def get_request(self, request):
        """Sets token-based auth headers."""
        if 'xml' in request.url:
            request.headers['employeesession'] = {
                'userId': self.user_id,
                'authToken': self.hash,
            }
        else:
            request.transport_user = self.user_id
            request.transport_password = self.hash
        return request

    def __repr__(self):
        return "EmployeeAuthentication(userId=%r,hash=%s)" % (self.user_id, self.hash)



getObject = {
    'accountId': 31111,
    'balancedTerminationFlag': False,
    'cooldown': 1800,
    'createDate': '2018-04-30T15:07:40-04:00',
    'desiredMemberCount': None,
    'id': 12222222,
    'lastActionDate': '2019-10-02T16:26:17-04:00',
    'loadBalancers': [],
    'maximumMemberCount': 6,
    'minimumMemberCount': 2,
    'modifyDate': '2019-10-03T17:16:47-04:00',
    'name': 'tests',
    'networkVlans': [
        {
            'networkVlan': {
                'accountId': 31111,
                'id': 2222222,
                'modifyDate': '2019-07-16T13:09:47-04:00',
                'networkSpace': 'PRIVATE',
                'primaryRouter': {
                    'hostname': 'bcr01a.sao01'
                },
                'primarySubnetId': 1172222,
                'vlanNumber': 1111
            },
            'networkVlanId': 2281111
        }
    ],
    'policies': [
        {
            'actions': [
                {
                    'amount': 1,
                    'createDate': '2019-09-26T18:30:22-04:00',
                    'deleteFlag': None,
                    'id': 611111,
                    'modifyDate': None,
                    'scalePolicy': None,
                    'scalePolicyId': 681111,
                    'scaleType': 'RELATIVE',
                    'typeId': 1
                }
            ],
            'cooldown': None,
            'createDate': '2019-09-26T18:30:14-04:00',
            'id': 680000,
            'name': 'prime-poly',
            'scaleActions': [
                {
                    'amount': 1,
                    'createDate': '2019-09-26T18:30:22-04:00',
                    'deleteFlag': None,
                    'id': 633333,
                    'modifyDate': None,
                    'scalePolicy': None,
                    'scalePolicyId': 680123,
                    'scaleType': 'RELATIVE',
                    'typeId': 1
                }
            ],
            'triggers': [
                {
                    'createDate': '2019-09-26T18:30:14-04:00',
                    'deleteFlag': None,
                    'id': 557111,
                    'modifyDate': None,
                    'scalePolicy': None,
                    'scalePolicyId': 680000,
                    'typeId': 3
                }
            ]
        }
    ],
    'regionalGroup': {
        'description': 'sa-bra-south-1',
        'id': 663,
        'locationGroupTypeId': 42,
        'locations': [
            {
                'id': 983497,
                'longName': 'Sao Paulo 1',
                'name': 'sao01',
                'statusId': 2
            }
        ],
        'name': 'sa-bra-south-1',
        'securityLevelId': None
    },
    'regionalGroupId': 663,
    'status': {
        'id': 1, 'keyName': 'ACTIVE', 'name': 'Active'
    },
    'suspendedFlag': False,
    'terminationPolicy': {
        'id': 2, 'keyName': 'NEWEST', 'name': 'Newest'
    },
    'terminationPolicyId': 2,
    'virtualGuestAssets': [],
    'virtualGuestMemberCount': 6,
    'virtualGuestMemberTemplate': {
        'accountId': 31111,
        'blockDevices': [
            {
                'bootableFlag': None,
                'createDate': None,
                'device': '0',
                'diskImage': {
                    'capacity': 25,
                    'createDate': None,
                    'id': None,
                    'modifyDate': None,
                    'parentId': None,
                    'storageRepositoryId': None,
                    'typeId': None},
                'diskImageId': None,
                'guestId': None,
                'hotPlugFlag': None,
                'id': None,
                'modifyDate': None,
                'statusId': None
            },
            {
                'bootableFlag': None,
                'createDate': None,
                'device': '2',
                'diskImage': {
                    'capacity': 10,
                    'createDate': None,
                    'id': None,
                    'modifyDate': None,
                    'parentId': None,
                    'storageRepositoryId': None,
                    'typeId': None},
                'diskImageId': None,
                'guestId': None,
                'hotPlugFlag': None,
                'id': None,
                'modifyDate': None,
                'statusId': None
            }
        ],
        'createDate': None,
        'datacenter': {
            'id': None,
            'name': 'sao01',
            'statusId': None
        },
        'dedicatedAccountHostOnlyFlag': None,
        'domain': 'tech-support.com',
        'hostname': 'testing',
        'hourlyBillingFlag': True,
        'id': None,
        'lastPowerStateId': None,
        'lastVerifiedDate': None,
        'localDiskFlag': False,
        'maxCpu': None,
        'maxMemory': 1024,
        'metricPollDate': None,
        'modifyDate': None,
        'networkComponents': [
            {
                'createDate': None,
                'guestId': None,
                'id': None,
                'maxSpeed': 100,
                'modifyDate': None,
                'networkId': None,
                'port': None,
                'speed': None
            }
        ],
        'operatingSystemReferenceCode': 'CENTOS_LATEST',
        'placementGroupId': None,
        'postInstallScriptUri': 'https://test.com/',
        'privateNetworkOnlyFlag': False,
        'provisionDate': None,
        'sshKeys': [
            {
                'createDate': None,
                'id': 490279,
                'modifyDate': None
            }
        ],
        'startCpus': 1,
        'statusId': None,
        'typeId': None},
    'virtualGuestMembers': [
        {
            'id': 3111111,
            'virtualGuest': {

                'domain': 'tech-support.com',
                'hostname': 'test',
                'provisionDate': '2019-09-27T14:29:53-04:00'
            }
        }
    ]
}

getVirtualGuestMembers = getObject['virtualGuestMembers']

scale = [
    {
        "accountId": 31111,
        "cooldown": 1800,
        "createDate": "2016-10-25T01:48:34+08:00",
        "id": 12222222,
        "maximumMemberCount": 5,
        "minimumMemberCount": 0,
        "name": "tests",
        "virtualGuest": {
            "accountId": 31111,
            "createDate": "2019-10-02T15:24:54-06:00",
            "billingItem": {
                "cancellationDate": "2019-10-02T08:34:21-06:00"}
        },
        "virtualGuestMemberTemplate": {
            "accountId": 31111,
            "domain": "sodg.com",
            "hostname": "testing",
            "maxMemory": 32768,
            "startCpus": 32,
            "blockDevices": [
                {
                    "device": "0",
                    "diskImage": {
                        "capacity": 25,
                    }
                }
            ],
            "datacenter": {
                "name": "sao01",
            },
            "hourlyBillingFlag": True,
            "operatingSystemReferenceCode": "CENTOS_LATEST",
            "privateNetworkOnlyFlag": True
        },
        "virtualGuestMemberCount": 0,
        "status": {
            "id": 1,
            "keyName": "ACTIVE",
            "name": "Active"
        },
        "virtualGuestAssets": [],
        "virtualGuestMembers": []
    },
    {
        "accountId": 31111,
        "cooldown": 1800,
        "createDate": "2018-04-24T04:22:00+08:00",
        "id": 224533333,
        "maximumMemberCount": 10,
        "minimumMemberCount": 0,
        "modifyDate": "2019-01-19T04:53:21+08:00",
        "name": "test-ajcb",
        "virtualGuest": {
            "accountId": 31111,
            "createDate": "2019-10-02T15:24:54-06:00",
            "billingItem": {
                "cancellationDate": "2019-10-02T08:34:21-06:00"}
        },
        "virtualGuestMemberTemplate": {
            "accountId": 31111,
            "domain": "test.local",
            "hostname": "autoscale-ajcb01",
            "id": None,
            "maxCpu": None,
            "maxMemory": 1024,
            "postInstallScriptUri": "http://test.com",
            "startCpus": 1,
            "blockDevices": [
                {
                    "device": "0",
                    "diskImage": {
                        "capacity": 25,
                    }
                }
            ],
            "datacenter": {
                "name": "seo01",
            },
            "hourlyBillingFlag": True,
            "operatingSystemReferenceCode": "CENTOS_7_64",
        },
        "virtualGuestMemberCount": 0,
        "status": {
            "id": 1,
            "keyName": "ACTIVE",
            "name": "Active"
        },
        "virtualGuestAssets": [],
        "virtualGuestMembers": []
    }
]

scaleTo = [
    {
        "accountId": 31111,
        "cooldown": 1800,
        "createDate": "2016-10-25T01:48:34+08:00",
        "id": 12222222,
        "lastActionDate": "2016-10-25T01:48:34+08:00",
        "maximumMemberCount": 5,
        "minimumMemberCount": 0,
        "name": "tests",
        "regionalGroupId": 663,
        "virtualGuest": {
        },
        "virtualGuestMemberTemplate": {
            "accountId": 31111,
            "domain": "sodg.com",
            "hostname": "testing",
            "id": None,
            "maxCpu": None,
            "maxMemory": 32768,
            "startCpus": 32,
            "datacenter": {
                "name": "sao01",
            },
            "hourlyBillingFlag": True,
            "operatingSystemReferenceCode": "CENTOS_LATEST",
            "privateNetworkOnlyFlag": True
        },
        "virtualGuestMemberCount": 0,
        "status": {
            "id": 1,
            "keyName": "ACTIVE",
            "name": "Active"
        },
        "virtualGuestAssets": [],
        "virtualGuestMembers": []
    },
    {
        "accountId": 31111,
        "cooldown": 1800,
        "createDate": "2018-04-24T04:22:00+08:00",
        "id": 224533333,
        "lastActionDate": "2019-01-19T04:53:18+08:00",
        "maximumMemberCount": 10,
        "minimumMemberCount": 0,
        "modifyDate": "2019-01-19T04:53:21+08:00",
        "name": "test-ajcb",
        "regionalGroupId": 1025,
        "virtualGuest": {
            "accountId": 31111,
            "createDate": "2019-10-02T15:24:54-06:00",
            "billingItem": {
                "cancellationDate": "2019-10-02T08:34:21-06:00"}
        },
        "virtualGuestMemberTemplate": {
            "accountId": 31111,
            "domain": "test.local",
            "hostname": "autoscale-ajcb01",
            "id": None,
            "maxCpu": None,
            "maxMemory": 1024,
            "postInstallScriptUri": "http://test.com",
            "startCpus": 1,
            "blockDevices": [
                {
                    "device": "0",
                    "diskImage": {
                        "capacity": 25,
                    }
                }
            ],
            "datacenter": {
                "name": "seo01",
            },
            "hourlyBillingFlag": True,
            "operatingSystemReferenceCode": "CENTOS_7_64",
        },
        "virtualGuestMemberCount": 0,
        "status": {
            "id": 1,
            "keyName": "ACTIVE",
            "name": "Active"
        },
        "virtualGuestAssets": [],
        "virtualGuestMembers": []
    },
]

getLogs = [
    {
        "createDate": "2019-10-03T04:26:11+08:00",
        "description": "Scaling group to 6 member(s) by adding 3 member(s) as manually requested",
        "id": 3821111,
        "scaleGroupId": 2252222,
        "scaleGroup": {
            "accountId": 31111,
            "cooldown": 1800,
            "createDate": "2018-05-01T03:07:40+08:00",
            "id": 2251111,
            "lastActionDate": "2019-10-03T04:26:17+08:00",
            "maximumMemberCount": 6,
            "minimumMemberCount": 2,
            "modifyDate": "2019-10-03T04:26:21+08:00",
            "name": "ajcb-autoscale11",
            "regionalGroupId": 663,
            "terminationPolicyId": 2,
            "virtualGuestMemberTemplate": {
                "accountId": 31111,
                "domain": "techsupport.com",
                "hostname": "ajcb-autoscale22",
                "maxMemory": 1024,
                "postInstallScriptUri": "https://pastebin.com/raw/62wrEKuW",
                "startCpus": 1,
                "blockDevices": [
                    {
                        "device": "0",
                        "diskImage": {
                            "capacity": 25,
                        }
                    },
                    {
                        "device": "2",
                        "diskImage": {
                            "capacity": 10,
                        }
                    }
                ],
                "datacenter": {
                    "name": "sao01",
                },
                "networkComponents": [
                    {
                        "maxSpeed": 100,
                    }
                ],
                "operatingSystemReferenceCode": "CENTOS_LATEST",
                "sshKeys": [
                    {
                        "id": 49111,
                    }
                ]
            },
            "logs": [
                {
                    "createDate": "2019-09-28T02:31:35+08:00",
                    "description": "Scaling group to 3 member(s) by removing -1 member(s) as manually requested",
                    "id": 3821111,
                    "scaleGroupId": 2251111,
                },
                {
                    "createDate": "2019-09-28T02:26:11+08:00",
                    "description": "Scaling group to 4 member(s) by adding 2 member(s) as manually requested",
                    "id": 38211111,
                    "scaleGroupId": 2251111,
                },
            ]
        }
    },
]

editObject = True
forceDeleteObject = True

createObject = {
    "accountId": 307608,
    "cooldown": 3600,
    "createDate": "2022-04-22T13:45:24-06:00",
    "id": 5446140,
    "lastActionDate": "2022-04-22T13:45:29-06:00",
    "maximumMemberCount": 5,
    "minimumMemberCount": 1,
    "name": "test22042022",
    "regionalGroupId": 4568,
    "suspendedFlag": False,
    "terminationPolicyId": 2,
    "virtualGuestMemberTemplate": {
        "accountId": 307608,
        "domain": "test.com",
        "hostname": "testvs",
        "maxMemory": 2048,
        "startCpus": 2,
        "blockDevices": [
            {
                "diskImage": {
                    "capacity": 100,
                }
            }
        ],
        "hourlyBillingFlag": True,
        "localDiskFlag": True,
        "networkComponents": [
            {
                "maxSpeed": 100,
            }
        ],
        "operatingSystemReferenceCode": "CENTOS_7_64",
        "userData": [
            {
                "value": "the userData"
            }
        ]
    },
    "virtualGuestMemberCount": 0,
    "networkVlans": [],
    "policies": [],
    "status": {
        "id": 1,
        "keyName": "ACTIVE",
        "name": "Active"
    },
    "virtualGuestAssets": [],
    "virtualGuestMembers": [
        {
            "createDate": "2022-04-22T13:45:29-06:00",
            "id": 123456,
            "scaleGroupId": 5446140,
            "virtualGuest": {
                "createDate": "2022-04-22T13:45:28-06:00",
                "deviceStatusId": 3,
                "domain": "test.com",
                "fullyQualifiedDomainName": "testvs-97e7.test.com",
                "hostname": "testvs-97e7",
                "id": 129911702,
                "maxCpu": 2,
                "maxCpuUnits": "CORE",
                "maxMemory": 2048,
                "startCpus": 2,
                "statusId": 1001,
                "typeId": 1,
                "uuid": "46e55f99-b412-4287-95b5-b8182b2fc924",
                "status": {
                    "keyName": "ACTIVE",
                    "name": "Active"
                }
            }
        }
    ]
}



"""List number of block storage volumes per datacenter."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting

DEFAULT_COLUMNS = [
    'Datacenter',
    'Count'
]


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.option('--datacenter', '-d', help='Datacenter shortname')
@click.option('--sortby', help='Column to sort by', default='Datacenter')
@environment.pass_env
def cli(env, sortby, datacenter):
    """List number of block storage volumes per datacenter."""
    block_manager = SoftLayer.BlockStorageManager(env.client)
    mask = "mask[serviceResource[datacenter[name]],"\
           "replicationPartners[serviceResource[datacenter[name]]]]"
    block_volumes = block_manager.list_block_volumes(datacenter=datacenter,
                                                     mask=mask)

    # cycle through all block volumes and count datacenter occurences.
    datacenters = {}
    for volume in block_volumes:
        service_resource = volume['serviceResource']
        if 'datacenter' in service_resource:
            datacenter_name = service_resource['datacenter']['name']
            if datacenter_name not in datacenters.keys():  # pylint: disable=consider-iterating-dictionary
                datacenters[datacenter_name] = 1
            else:
                datacenters[datacenter_name] += 1

    table = formatting.KeyValueTable(DEFAULT_COLUMNS)
    table.sortby = sortby
    for key, value in datacenters.items():
        table.add_row([key, value])
    env.fout(table)



"""Adds an attachment to an existing ticket."""
# :license: MIT, see LICENSE for more details.

import os

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--path', help="The path of the attachment to be uploaded")
@click.option('--name', help="The name of the attachment shown in the ticket")
@environment.pass_env
def cli(env, identifier, path, name):
    """Adds an attachment to an existing ticket."""
    mgr = SoftLayer.TicketManager(env.client)

    ticket_id = helpers.resolve_id(mgr.resolve_ids, identifier, 'ticket')

    if path is None:
        raise exceptions.ArgumentError("Missing argument --path")

    if not os.path.exists(path):
        raise exceptions.ArgumentError("%s not exist" % path)

    if name is None:
        name = os.path.basename(path)

    attached_file = mgr.upload_attachment(ticket_id=ticket_id,
                                          file_path=path,
                                          file_name=name)
    env.fout("File attached: \n%s" % attached_file)



"""
    SoftLayer.tags
    ~~~~~~~~~~~~~~
    Tag Manager

    :license: MIT, see LICENSE for more details.
"""
import re

from SoftLayer.exceptions import SoftLayerAPIError


class TagManager(object):
    """Manager for Tag functions."""

    def __init__(self, client):
        self.client = client

    def list_tags(self, mask=None):
        """Returns a list of all tags for the Current User

        :param str mask: Object mask to use if you do not want the default.
        """
        if mask is None:
            mask = "mask[id,name,referenceCount]"
        unattached = self.get_unattached_tags(mask)
        attached = self.get_attached_tags(mask)
        return {'attached': attached, 'unattached': unattached}

    def get_unattached_tags(self, mask=None):
        """Calls SoftLayer_Tag::getUnattachedTagsForCurrentUser()

        :params string mask: Mask to use.
        """
        return self.client.call('SoftLayer_Tag', 'getUnattachedTagsForCurrentUser',
                                mask=mask, iter=True)

    def get_attached_tags(self, mask=None):
        """Calls SoftLayer_Tag::getAttachedTagsForCurrentUser()

        :params string mask: Mask to use.
        """
        return self.client.call('SoftLayer_Tag', 'getAttachedTagsForCurrentUser',
                                mask=mask, iter=True)

    def get_tag_references(self, tag_id, mask=None):
        """Calls SoftLayer_Tag::getReferences(id=tag_id)

        :params int tag_id: Tag id to get references from
        :params string mask: Mask to use.
        """
        if mask is None:
            mask = "mask[tagType]"
        return self.client.call('SoftLayer_Tag', 'getReferences', id=tag_id, mask=mask, iter=True)

    def get_tag(self, tag_id, mask=None):
        """Calls SoftLayer_Tag::getObject(id=tag_id)

        :params int tag_id: Tag id to get object from
        :params string mask: Mask to use.
        """
        if mask is None:
            mask = "mask[id,name]"
        result = self.client.call('SoftLayer_Tag', 'getObject', id=tag_id, mask=mask)
        return result

    def get_tag_by_name(self, tag_name, mask=None):
        """Calls SoftLayer_Tag::getTagByTagName(tag_name)

        :params string tag_name: Tag name to get object from
        :params string mask: Mask to use.
        """
        if mask is None:
            mask = "mask[id,name]"
        result = self.client.call('SoftLayer_Tag', 'getTagByTagName', tag_name, mask=mask)
        return result

    def reference_lookup(self, resource_table_id, tag_type):
        """Returns the SoftLayer Service for the corresponding type

        :param int resource_table_id: Tag_Reference::resourceTableId
        :param string tag_type: Tag_Reference->tagType->keyName

        From  SoftLayer_Tag::getAllTagTypes()

        ::

            |Type                             |Service |
            | -----------------------------   | ------ |
            |Hardware                         |HARDWARE|
            |CCI                              |GUEST|
            |Account Document                 |ACCOUNT_DOCUMENT|
            |Ticket                           |TICKET|
            |Vlan Firewall                    |NETWORK_VLAN_FIREWALL|
            |Contract                         |CONTRACT|
            |Image Template                   |IMAGE_TEMPLATE|
            |Application Delivery Controller  |APPLICATION_DELIVERY_CONTROLLER|
            |Vlan                             |NETWORK_VLAN|
            |Dedicated Host                   |DEDICATED_HOST|
        """
        service = self.type_to_service(tag_type)
        if service is None:
            raise SoftLayerAPIError(404, f"Unable to lookup {tag_type} types")
        return self.client.call(service, 'getObject', id=resource_table_id)

    def delete_tag(self, name):
        """Calls SoftLayer_Tag::deleteTag

        :param string name: tag name to delete
        """
        return self.client.call('SoftLayer_Tag', 'deleteTag', name)

    def set_tags(self, tags, key_name, resource_id):
        """Calls SoftLayer_Tag::setTags()

        :param string tags: List of tags.
        :param string key_name: Key name of a tag type.
        :param int resource_id: ID of the object being tagged.
        """
        return self.client.call('SoftLayer_Tag', 'setTags', tags, key_name, resource_id)

    def get_all_tag_types(self):
        """Calls SoftLayer_Tag::getAllTagTypes()"""
        types = self.client.call('SoftLayer_Tag', 'getAllTagTypes')
        useable_types = []
        for tag_type in types:
            service = self.type_to_service(tag_type['keyName'])
            # Mostly just to remove the types that are not user taggable.
            if service is not None:
                temp_type = tag_type
                temp_type['service'] = service
                useable_types.append(temp_type)
        return useable_types

    def taggable_by_type(self, tag_type):
        """Returns a list of resources that can be tagged, that are of the given type

        :param string tag_type: Key name of a tag type. See SoftLayer_Tag::getAllTagTypes
        """
        service = self.type_to_service(tag_type)
        search_term = f"_objectType:SoftLayer_{service}"
        if tag_type == 'TICKET':
            search_term = f"{search_term} status.name: open"
        elif tag_type == 'IMAGE_TEMPLATE':
            mask = "mask[id,accountId,name,globalIdentifier,parentId,publicFlag,flexImageFlag,imageType]"
            resources = self.client.call('SoftLayer_Account', 'getPrivateBlockDeviceTemplateGroups',
                                         mask=mask, iter=True)
            to_return = []
            # Fake search result output
            for resource in resources:
                to_return.append({'resourceType': service, 'resource': resource})
            return to_return
        elif tag_type == 'NETWORK_SUBNET':
            resources = self.client.call('SoftLayer_Account', 'getSubnets', iter=True)
            to_return = []
            # Fake search result output
            for resource in resources:
                to_return.append({'resourceType': service, 'resource': resource})
            return to_return
        resources = self.client.call('SoftLayer_Search', 'advancedSearch', search_term, iter=True)
        return resources

    @staticmethod
    def type_to_service(tag_type):
        """Returns the SoftLayer service for the given tag_type"""
        service = None
        if tag_type in ['ACCOUNT_DOCUMENT', 'CONTRACT']:
            return None

        if tag_type == 'APPLICATION_DELIVERY_CONTROLLER':
            service = 'Network_Application_Delivery_Controller'
        elif tag_type == 'GUEST':
            service = 'Virtual_Guest'
        elif tag_type == 'DEDICATED_HOST':
            service = 'Virtual_DedicatedHost'
        elif tag_type == 'IMAGE_TEMPLATE':
            service = 'Virtual_Guest_Block_Device_Template_Group'
        else:

            tag_type = tag_type.lower()
            # Sets the First letter, and any letter preceeded by a '_' to uppercase
            # HARDWARE -> Hardware, NETWORK_VLAN -> Network_Vlan for example.
            service = re.sub(r'(^[a-z]|\_[a-z])', lambda x: x.group().upper(), tag_type)
        return service

    @staticmethod
    def get_resource_name(resource, tag_type):
        """Returns a string that names a resource

        :param dict resource: A SoftLayer datatype for the given tag_type
        :param string tag_type: Key name for the tag_type
        """
        if tag_type == 'NETWORK_VLAN_FIREWALL':
            return resource.get('primaryIpAddress')
        elif tag_type == 'NETWORK_VLAN':
            return f"{resource.get('vlanNumber')} ({resource.get('name')})"
        elif tag_type == 'IMAGE_TEMPLATE' or tag_type == 'APPLICATION_DELIVERY_CONTROLLER':
            return resource.get('name')
        elif tag_type == 'TICKET':
            return resource.get('subjet')
        elif tag_type == 'NETWORK_SUBNET':
            return resource.get('networkIdentifier')
        else:
            return resource.get('fullyQualifiedDomainName')

    # @staticmethod
    # def type_to_datatype(tag_type):
    #     """Returns the SoftLayer datatye for the given tag_type"""
    #     datatye = None
    #     if tag_type in ['ACCOUNT_DOCUMENT', 'CONTRACT']:
    #         return None

    #     if tag_type == 'APPLICATION_DELIVERY_CONTROLLER':
    #         datatye = 'adcLoadBalancers'
    #     elif tag_type == 'GUEST':
    #         datatye = 'virtualGuests'
    #     elif tag_type == 'DEDICATED_HOST':
    #         datatye = 'dedicatedHosts'
    #     elif tag_type == 'HARDWARE':
    #         datatye = 'hardware'
    #     elif tag_type == 'TICKET':
    #         datatye = 'openTickets'
    #     elif tag_type == 'NETWORK_SUBNET':
    #         datatye = 'subnets'
    #     elif tag_type == 'NETWORK_VLAN':
    #         datatye = 'networkVlans'
    #     elif tag_type == 'NETWORK_VLAN_FIREWALL':
    #         datatye = 'networkVlans'
    #     elif tag_type == 'IMAGE_TEMPLATE':
    #         datatye = 'blockDeviceTemplateGroups'

    #     return datatye



"""Licenses order options for a given VMware licenses."""
# :license: MIT, see LICENSE for more details.

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers.license import LicensesManager
from SoftLayer import utils


@click.command(cls=SLCommand)
@environment.pass_env
def cli(env):
    """Server order options for a given chassis."""

    licenses_manager = LicensesManager(env.client)

    options = licenses_manager.get_create_options()

    table = formatting.Table(['Id', 'description', 'keyName', 'capacity', 'recurringFee'])
    for item in options:
        fee = 0
        if len(item.get('prices', [])) > 0:
            fee = item.get('prices')[0].get('recurringFee', 0)
        table.add_row([item.get('id'),
                       utils.trim_to(item.get('description'), 40),
                       item.get('keyName'),
                       item.get('capacity'),
                       fee])

    env.fout(table)



getObject = {
    'bandwidthAllotmentTypeId': 2,
    'createDate': '2016-07-25T08:31:17-07:00',
    'id': 123456,
    'locationGroupId': 262,
    'name': 'MexRegion',
    'serviceProviderId': 1,
    'bareMetalInstanceCount': 0,
    'hardwareCount': 2,
    'virtualGuestCount': 0,
    'activeDetails': [
        {
            'allocationId': 48293300,
            'bandwidthAllotmentId': 309961,
            'effectiveDate': '2022-02-04T00:00:00-06:00',
            'id': 48882086,
            'serviceProviderId': 1,
            'allocation': {
                'amount': '5000',
                'id': 48293300,
            }
        },
        {
            'allocationId': 48293302,
            'bandwidthAllotmentId': 309961,
            'effectiveDate': '2022-02-04T00:00:00-06:00',
            'id': 48882088,
            'serviceProviderId': 1,
            'allocation': {
                'amount': '5000',
                'id': 48293302,
            }
        }
    ],
    'bareMetalInstances': [],
    'billingCyclePublicBandwidthUsage': {
        'amountIn': '.23642',
        'amountOut': '.05475',
        'bandwidthUsageDetailTypeId': '1',
        'trackingObject': {
            'id': 258963,
            'resourceTableId': 309961,
            'startDate': '2021-03-10T11:04:56-06:00',
        }
    },
    'hardware': [
        {
            'domain': 'test.com',
            'fullyQualifiedDomainName': 'testpooling.test.com',
            'hardwareStatusId': 5,
            'hostname': 'testpooling',
            'id': 36589,
            'manufacturerSerialNumber': 'J122Y7N',
            'provisionDate': '2022-01-24T15:17:03-06:00',
            'serialNumber': 'SL018EA8',
            'serviceProviderId': 1,
            'bandwidthAllotmentDetail': {
                'allocationId': 48293302,
                'bandwidthAllotmentId': 309961,
                'effectiveDate': '2022-02-04T00:00:00-06:00',
                'id': 48882088,
                'allocation': {
                    'amount': '5000',
                    'id': 48293302,
                }
            },
            'globalIdentifier': '36e63026-5fa1-456d-a04f-adf34e60e2f4',
            'hardwareStatus': {
                'id': 5,
                'status': 'ACTIVE'
            },
            'networkManagementIpAddress': '10.130.97.247',
            'outboundBandwidthUsage': '.02594',
            'primaryBackendIpAddress': '10.130.97.227',
            'primaryIpAddress': '169.57.4.70',
            'privateIpAddress': '10.130.97.227'
        },
        {
            'domain': 'testtest.com',
            'fullyQualifiedDomainName': 'testpooling2.test.com',
            'hardwareStatusId': 5,
            'hostname': 'testpooling2',
            'id': 25478,
            'manufacturerSerialNumber': 'J12935M',
            'notes': '',
            'provisionDate': '2022-01-24T15:44:20-06:00',
            'serialNumber': 'SL01HIIB',
            'serviceProviderId': 1,
            'bandwidthAllotmentDetail': {
                'allocationId': 48293300,
                'bandwidthAllotmentId': 309961,
                'effectiveDate': '2022-02-04T00:00:00-06:00',
                'id': 48882086,
                'serviceProviderId': 1,
                'allocation': {
                    'amount': '5000',
                    'id': 478965,
                }
            },
            'globalIdentifier': '6ea407bd-9c07-4129-9103-9fda8a9e7028',
            'hardwareStatus': {
                'id': 5,
                'status': 'ACTIVE'
            },
            'networkManagementIpAddress': '10.130.97.252',
            'outboundBandwidthUsage': '.02884',
            'primaryBackendIpAddress': '10.130.97.248',
            'primaryIpAddress': '169.57.4.73',
            'privateIpAddress': '10.130.97.248'
        }
    ],
    'inboundPublicBandwidthUsage': '.23642',
    'projectedPublicBandwidthUsage': 0.43,
    'virtualGuests': [{
        'createDate': '2021-06-09T13:49:28-07:00',
        'deviceStatusId': 8,
        'domain': 'cgallo.com',
        'fullyQualifiedDomainName': 'KVM-Test.test.com',
        'hostname': 'KVM-Test',
        'id': 3578963,
        'maxCpu': 2,
        'maxCpuUnits': 'CORE',
        'maxMemory': 4096,
        'startCpus': 2,
        'statusId': 1001,
        'typeId': 1,
        'uuid': '15951561-6171-0dfc-f3d2-be039e51cc10',
        'bandwidthAllotmentDetail': {
            'allocationId': 45907006,
            'bandwidthAllotmentId': 138442,
            'effectiveDate': '2021-06-09T13:49:31-07:00',
            'id': 46467342,
            'serviceProviderId': 1,
            'allocation': {
                'amount': '0',
                'id': 45907006,
            }
        },
        'globalIdentifier': 'a245a7dd-acd1-4d1a-9356-cc1ac6b55b98',
        'outboundPublicBandwidthUsage': '.02845',
        'primaryBackendIpAddress': '10.208.73.53',
        'primaryIpAddress': '169.48.96.27',
        'status': {
            'keyName': 'ACTIVE',
            'name': 'Active'
        }
    }]
}
createObject = {
    "accountId": 654321,
    "bandwidthAllotmentTypeId": 2,
    "createDate": "2023-05-31T07:22:27-06:00",
    "endDate": None,
    "id": 123456789,
    "locationGroupId": 2,
    "name": "NewRegion",
    "serviceProviderId": 1
}

editObject = True

requestVdrCancellation = True



"""
    SoftLayer.event_log
    ~~~~~~~~~~~~~~~~~~~
    Network Manager/helpers

    :license: MIT, see LICENSE for more details.
"""

from SoftLayer import utils


class EventLogManager(object):
    """Provides an interface for the SoftLayer Event Log Service.

    See product information here:
    http://sldn.softlayer.com/reference/services/SoftLayer_Event_Log
    """

    def __init__(self, client):
        self.client = client
        self.event_log = client['Event_Log']

    def get_event_logs(self, request_filter=None, log_limit=20, iterator=True):
        """Returns a list of event logs

        Example::

            event_mgr = SoftLayer.EventLogManager(env.client)
            request_filter = event_mgr.build_filter(date_min="01/01/2019", date_max="02/01/2019")
            logs = event_mgr.get_event_logs(request_filter)
            for log in logs:
                print("Event Name: {}".format(log['eventName']))


        :param dict request_filter: filter dict
        :param int log_limit: number of results to get in one API call
        :param bool iterator: False will only make one API call for log_limit results.
            True will keep making API calls until all logs have been retreived. There may be a lot of these.
        :returns: List of event logs. If iterator=True, will return a python generator object instead.
        """
        if iterator:
            # Call iter_call directly as this returns the actual generator
            return self.client.iter_call('Event_Log', 'getAllObjects', filter=request_filter, limit=log_limit)
        return self.client.call('Event_Log', 'getAllObjects', filter=request_filter, limit=log_limit)

    def get_event_log_types(self):
        """Returns a list of event log types

        :returns: List of event log types
        """
        results = self.event_log.getAllEventObjectNames()
        return results

    @staticmethod
    def build_filter(date_min=None, date_max=None, obj_event=None, obj_id=None, obj_type=None, utc_offset=None):
        """Returns a query filter that can be passed into EventLogManager.get_event_logs

        :param string date_min: Lower bound date in MM/DD/YYYY format
        :param string date_max: Upper bound date in MM/DD/YYYY format
        :param string obj_event: The name of the events we want to filter by
        :param int obj_id: The id of the event we want to filter by
        :param string obj_type: The type of event we want to filter by
        :param string utc_offset: The UTC offset we want to use when converting date_min and date_max.
            (default '+0000')

        :returns: dict: The generated query filter
        """
        request_filter = {}
        request_filter['traceId'] = utils.query_filter_orderby()

        if date_min and date_max:
            request_filter['eventCreateDate'] = utils.event_log_filter_between_date(date_min, date_max, utc_offset)
        else:
            if date_min:
                request_filter['eventCreateDate'] = utils.event_log_filter_greater_than_date(date_min, utc_offset)
            elif date_max:
                request_filter['eventCreateDate'] = utils.event_log_filter_less_than_date(date_max, utc_offset)

        if obj_event:
            request_filter['eventName'] = {'operation': obj_event}

        if obj_id:
            request_filter['objectId'] = {'operation': obj_id}

        if obj_type:
            request_filter['objectName'] = {'operation': obj_type}

        return request_filter



listDomainMappings = [
    {
        "cacheKeyQueryRule": "include-all",
        "cname": "cdnakauuiet7s6u6.cdnedge.bluemix.net",
        "createDate": "2020-09-29T15:19:01-06:00",
        "domain": "test.example.com",
        "header": "test.example.com",
        "httpPort": 80,
        "originHost": "1.1.1.1",
        "originType": "HOST_SERVER",
        "path": "/",
        "protocol": "HTTP",
        "status": "CNAME_CONFIGURATION",
        "uniqueId": "11223344",
        "vendorName": "akamai"
    }
]

listDomainMappingByUniqueId = [
    {
        "cname": "cdnakauuiet7s6u6.cdnedge.bluemix.net",
        "performanceConfiguration": "Large file optimization",
        "domain": "test.example.com",
        "header": "test.example.com",
        "httpPort": 80,
        "originHost": "1.1.1.1",
        "originType": "HOST_SERVER",
        "path": "/",
        "protocol": "HTTP",
        "status": "CNAME_CONFIGURATION",
        "uniqueId": "11223344",
        "vendorName": "akamai"
    }
]

updateDomainMapping = [
    {
        "createDate": "2021-02-09T19:32:29-06:00",
        "originType": "HOST_SERVER",
        "path": "/*",
        "performanceConfiguration": "Large file optimization",
        "protocol": "HTTP",
        "respectHeaders": True,
        "uniqueId": "11223344",
        "vendorName": "akamai",
        "header": "www.test.com",
        "httpPort": 83,
        "cname": "cdn.test.cloud",
        "originHost": "1.1.1.1",
        "cacheKeyQueryRule": "include: test"
    }
]

deleteDomainMapping = [
    {
        "akamaiCname": "wildcard.appdomain.mdc.edgekey.net",
        "certificateType": "NO_CERT",
        "cname": "cdnakayq9fye4t88239.cdn.appdomain.cloud",
        "createDate": "2023-03-29T07:33:42-06:00",
        "domain": "www.test.com",
        "header": "www.header.com",
        "httpPort": 80,
        "httpsPort": None,
        "modifyDate": "2023-03-29T07:33:46-06:00",
        "originHost": "67.228.227.82",
        "originType": "HOST_SERVER",
        "path": "/path/*",
        "performanceConfiguration": "General web delivery",
        "protocol": "HTTP",
        "respectHeaders": False,
        "serveStale": True,
        "status": "DELETING",
        "uniqueId": "303727924488685",
        "vendorName": "akamai"
    }
]
createDomainMapping = [
    {
        "bucketName": "test-bucket-name",
        "akamaiCname": "wildcard.appdomain.mdc.edgekey.net",
        "cacheKeyQueryRule": "include-all",
        "certificateType": "WILDCARD_CERT",
        "cname": "test.cdn.appdomain.cloud",
        "createDate": "2020-09-29T15:19:01-06:00",
        "domain": "test.com",
        "header": "header.test.com",
        "httpPort": 80,
        "httpsPort": None,
        "modifyDate": "2021-06-24T09:02:22-06:00",
        "originHost": "10.32.12.125",
        "originType": "HOST_SERVER",
        "path": "/*",
        "performanceConfiguration": "General web delivery",
        "protocol": "HTTP",
        "respectHeaders": True,
        "serveStale": True,
        "status": "CNAME_CONFIGURATION",
        "uniqueId": "354034879028850",
        "vendorName": "akamai"
    }
]



"""Display details for a specified volume."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers
from SoftLayer import utils


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('volume_id')
@environment.pass_env
def cli(env, volume_id):
    """Display details for a specified volume."""
    block_manager = SoftLayer.BlockStorageManager(env.client)
    block_volume_id = helpers.resolve_id(block_manager.resolve_ids, volume_id, 'Block Volume')
    block_volume = block_manager.get_block_volume_details(block_volume_id)
    block_volume = utils.NestedDict(block_volume)

    table = formatting.KeyValueTable(['Name', 'Value'])
    table.align['Name'] = 'r'
    table.align['Value'] = 'l'

    capacity = '0'
    if block_volume['capacityGb'] != '':
        capacity = "%iGB" % block_volume['capacityGb']
    storage_type = block_volume['storageType']['keyName'].split('_').pop(0)
    table.add_row(['ID', block_volume['id']])
    table.add_row(['Username', block_volume['username']])
    table.add_row(['Type', storage_type])
    table.add_row(['Capacity (GB)', capacity])
    table.add_row(['LUN Id', block_volume['lunId']])

    if block_volume.get('provisionedIops'):
        table.add_row(['IOPs', block_volume['provisionedIops']])

    if block_volume.get('storageTierLevel'):
        table.add_row(['Endurance Tier', block_volume['storageTierLevel']])

    table.add_row(['Data Center', block_volume['serviceResource']['datacenter']['name']])
    table.add_row(['Target IP', block_volume['serviceResourceBackendIpAddress']])

    if block_volume['snapshotCapacityGb']:
        table.add_row(['Snapshot Capacity (GB)', block_volume['snapshotCapacityGb']])
        if 'snapshotSizeBytes' in block_volume['parentVolume']:
            table.add_row(['Snapshot Used (Bytes)', block_volume['parentVolume']['snapshotSizeBytes']])

    table.add_row(['# of Active Transactions', block_volume['activeTransactionCount']])

    if block_volume['activeTransactions']:
        for trans in block_volume['activeTransactions']:
            if 'transactionStatus' in trans and 'friendlyName' in trans['transactionStatus']:
                table.add_row(['Ongoing Transaction', trans['transactionStatus']['friendlyName']])

    table.add_row(['Replicant Count', block_volume.get('replicationPartnerCount', 0)])

    if block_volume['replicationPartnerCount'] > 0:
        # This if/else temporarily handles a bug in which the SL API
        # returns a string or object for 'replicationStatus'; it seems that
        # the type is string for File volumes and object for Block volumes
        if 'message' in block_volume['replicationStatus']:
            table.add_row(['Replication Status', block_volume['replicationStatus']['message']])
        else:
            table.add_row(['Replication Status', block_volume['replicationStatus']])

        replicant_table = formatting.Table(['Id', 'Username', 'Target', 'Location', 'Schedule'])
        replicant_table.align['Name'] = 'r'
        replicant_table.align['Value'] = 'l'
        for replicant in block_volume['replicationPartners']:
            replicant_table.add_row([
                replicant.get('id'),
                utils.lookup(replicant, 'username'),
                utils.lookup(replicant, 'serviceResourceBackendIpAddress'),
                utils.lookup(replicant, 'serviceResource', 'datacenter', 'name'),
                utils.lookup(replicant, 'replicationSchedule', 'type', 'keyname')
            ])
        table.add_row(['Replicant Volumes', replicant_table])

    if block_volume.get('originalVolumeSize'):
        original_volume_info = formatting.Table(['Property', 'Value'])
        original_volume_info.add_row(['Original Volume Size', block_volume['originalVolumeSize']])
        if block_volume.get('originalVolumeName'):
            original_volume_info.add_row(['Original Volume Name', block_volume['originalVolumeName']])
        if block_volume.get('originalSnapshotName'):
            original_volume_info.add_row(['Original Snapshot Name', block_volume['originalSnapshotName']])
        table.add_row(['Original Volume Properties', original_volume_info])

    notes = f"{block_volume.get('notes', '')}"
    table.add_row(['Notes', notes])

    env.fout(table)
